# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from qinling.db.sqlalchemy import model_base
from qinling.db.sqlalchemy import types as st
from qinling.utils import common


class Runtime(model_base.QinlingSecureModelBase):
    __tablename__ = 'runtimes'

    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    image = sa.Column(sa.String(255), nullable=False)
    status = sa.Column(sa.String(32), nullable=False)
    is_public = sa.Column(sa.BOOLEAN, default=True)
    trusted = sa.Column(sa.BOOLEAN, default=True)


class Function(model_base.QinlingSecureModelBase):
    __tablename__ = 'functions'

    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255))
    runtime_id = sa.Column(
        sa.String(36), sa.ForeignKey(Runtime.id), nullable=True
    )
    cpu = sa.Column(sa.Integer, default=0)
    memory_size = sa.Column(sa.Integer, default=0)
    timeout = sa.Column(sa.Integer)
    code = sa.Column(st.JsonLongDictType(), nullable=False)
    entry = sa.Column(sa.String(80), nullable=True)
    count = sa.Column(sa.Integer, default=0)
    trust_id = sa.Column(sa.String(80))
    latest_version = sa.Column(sa.Integer, default=0)


class Execution(model_base.QinlingSecureModelBase):
    __tablename__ = 'executions'

    function_alias = sa.Column(sa.String(255), nullable=True)
    function_id = sa.Column(sa.String(36), nullable=True)
    function_version = sa.Column(sa.Integer, default=0)
    status = sa.Column(sa.String(32), nullable=False)
    sync = sa.Column(sa.BOOLEAN, default=True)
    input = sa.Column(st.JsonLongDictType())
    result = sa.Column(st.JsonLongDictType())
    description = sa.Column(sa.String(255))
    logs = sa.Column(sa.Text(), nullable=True)


class Job(model_base.QinlingSecureModelBase):
    __tablename__ = 'jobs'

    name = sa.Column(sa.String(255), nullable=True)
    pattern = sa.Column(sa.String(32), nullable=True)
    status = sa.Column(sa.String(32), nullable=False)
    first_execution_time = sa.Column(sa.DateTime, nullable=True)
    next_execution_time = sa.Column(sa.DateTime, nullable=False)
    count = sa.Column(sa.Integer)
    function_id = sa.Column(
        sa.String(36),
        sa.ForeignKey(Function.id),
        nullable=True
    )
    function = relationship('Function', back_populates="jobs")
    function_input = sa.Column(sa.String(255), nullable=True)
    function_version = sa.Column(sa.Integer, default=0)
    function_alias = sa.Column(sa.String(255), nullable=True)

    def to_dict(self):
        d = super(Job, self).to_dict()
        common.datetime_to_str(d, 'first_execution_time')
        common.datetime_to_str(d, 'next_execution_time')
        return d


class Webhook(model_base.QinlingSecureModelBase):
    __tablename__ = 'webhooks'

    function_alias = sa.Column(sa.String(255), nullable=True)
    function_id = sa.Column(
        sa.String(36),
        sa.ForeignKey(Function.id),
        nullable=True
    )
    function_version = sa.Column(sa.Integer, nullable=True)
    description = sa.Column(sa.String(255))


class FunctionVersion(model_base.QinlingSecureModelBase):
    __tablename__ = 'function_versions'

    __table_args__ = (
        sa.UniqueConstraint('project_id', 'function_id', 'version_number'),
        sa.Index(
            '%s_project_id_function_id_version_number' % __tablename__,
            'project_id',
            'function_id',
            'version_number'
        )
    )

    function_id = sa.Column(
        sa.String(36),
        sa.ForeignKey(Function.id, ondelete='CASCADE')
    )
    function = relationship('Function', back_populates="versions")
    description = sa.Column(sa.String(255), nullable=True)
    version_number = sa.Column(sa.Integer, default=0)
    count = sa.Column(sa.Integer, default=0)


class FunctionAlias(model_base.QinlingSecureModelBase):
    __tablename__ = 'function_aliases'

    __table_args__ = (
        sa.UniqueConstraint('project_id', 'function_id', 'function_version'),
        sa.Index(
            '%s_project_id_function_id_function_version' % __tablename__,
            'project_id',
            'function_id',
            'function_version'
        ),
        sa.UniqueConstraint('project_id', 'name'),
        sa.Index(
            '%s_project_id_name' % __tablename__,
            'project_id',
            'name'
        )
    )

    function_id = sa.Column(
        sa.String(36),
        sa.ForeignKey(Function.id)
    )
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.String(255), nullable=True)
    function_version = sa.Column(sa.Integer, default=0)


# Only get running jobs
Function.jobs = relationship(
    "Job",
    back_populates="function",
    primaryjoin=(
        "and_(Function.id==Job.function_id, "
        "~Job.status.in_(['done', 'cancelled']))"
    )
)
Function.webhooks = relationship("Webhook", uselist=True, backref="function")
Function.versions = relationship(
    "FunctionVersion",
    order_by="FunctionVersion.version_number",
    uselist=True,
    lazy='select',
    cascade="all, delete-orphan"
)
Function.aliases = relationship(
    "FunctionAlias",
    uselist=True,
    backref="function"
)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import etcd3gw
from oslo_config import cfg

CONF = cfg.CONF
CLIENT = None


def get_client(conf=None):
    global CLIENT
    conf = conf or CONF

    if not CLIENT:
        if conf.etcd.protocol == "https":
            CLIENT = etcd3gw.client(host=conf.etcd.host,
                                    port=conf.etcd.port,
                                    protocol=conf.etcd.protocol,
                                    ca_cert=conf.etcd.ca_cert,
                                    cert_cert=conf.etcd.cert_file,
                                    cert_key=conf.etcd.cert_key)
        else:
            CLIENT = etcd3gw.client(host=conf.etcd.host,
                                    port=conf.etcd.port,
                                    protocol=conf.etcd.protocol)

    return CLIENT


def get_worker_lock(function_id, version=0):
    client = get_client()
    lock_id = "function_worker_%s_%s" % (function_id, version)
    return client.lock(id=lock_id)


def get_function_version_lock(function_id):
    client = get_client()
    lock_id = "function_version_%s" % function_id
    return client.lock(id=lock_id)


def create_worker(function_id, worker, version=0):
    """Create the worker info in etcd.

    The worker parameter is assumed to be unique.
    """
    # NOTE(huntxu): for the kubernetes orchestrator, which is the only
    # available orchestrator at the moment, the value of the worker param
    # is the name of the pod so it is unique.
    client = get_client()
    client.create(
        '%s_%s/worker_%s' % (function_id, version, worker),
        worker
    )


def delete_worker(function_id, worker, version=0):
    client = get_client()
    client.delete('%s_%s/worker_%s' % (function_id, version, worker))


def get_workers(function_id, version=0):
    client = get_client()
    values = client.get_prefix("%s_%s/worker" % (function_id, version))
    workers = [w[0] for w in values]
    return workers


def delete_function(function_id, version=0):
    client = get_client()
    client.delete_prefix("%s_%s" % (function_id, version))


def create_service_url(function_id, url, version=0):
    client = get_client()
    client.create('%s_%s/service_url' % (function_id, version), url)


def get_service_url(function_id, version=0):
    client = get_client()
    values = client.get('%s_%s/service_url' % (function_id, version))
    return None if not values else values[0]



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import functools
import hashlib
import pdb
import sys
import warnings

from oslo_utils import uuidutils
import six

from qinling import exceptions as exc
from qinling import version


def print_server_info(service):
    QINLING_TITLE = r"""
                                    /^L_      ,."\
               /~\       __       /~    \   ./    \
              /   _\   _/  \     /T~\|~\_\ / \_  /~|          _^
            / \ /W  \ / V^\/X  /~         T  . \/   \    ,v-./
     ,'`-. /~   ^     H  ,  . \/    ;   .   \      `. \-'   /
         M      ~     | . ;  /         ,  _   :  .    ~\_,-'
        /    ~    .    \    /   :                   '   \   ,/`
       I o. ^    oP     '98b         -      _  9.`       `\9b.
     8oO888.  oO888P  d888b9bo. .8o 888o.       8bo.  o     988o.
     88888888888888888888888888bo.98888888bo.    98888bo. .d888P
     88888888888888888888888888888888888888888888888888888888888
                         _          __   _
                 ___ _  (_)  ___   / /  (_)  ___   ___ _
                / _ `/ / /  / _ \ / /  / /  / _ \ / _ `/
                \_, / /_/  /_//_//_/  /_/  /_//_/ \_, /
                 /_/                             /___/

    Function as a Service in OpenStack, version: %s
    """ % version.version_string()

    print(QINLING_TITLE)
    print('Launching server components %s...' % service)


def get_properly_ordered_parameters():
    """Orders launch parameters in the right order.

    In oslo it's important the order of the launch parameters.
    if --config-file came after the command line parameters the command
    line parameters are ignored.
    So to make user command line parameters are never ignored this method
    moves --config-file to be always first.
    """
    args = sys.argv[1:]

    for arg in sys.argv[1:]:
        if arg == '--config-file' or arg.startswith('--config-file='):
            if "=" in arg:
                conf_file_value = arg.split("=", 1)[1]
            else:
                conf_file_value = args[args.index(arg) + 1]
                args.remove(conf_file_value)
            args.remove(arg)
            args.insert(0, "--config-file")
            args.insert(1, conf_file_value)

    return args


def convert_dict_to_string(d):
    temp_list = ['%s=%s' % (k, v) for k, v in d.items()]

    return ','.join(temp_list)


def datetime_to_str(dct, attr_name):
    """Convert datetime object in dict to string."""
    if (dct.get(attr_name) is not None and
            not isinstance(dct.get(attr_name), six.string_types)):
        dct[attr_name] = dct[attr_name].strftime('%Y-%m-%dT%H:%M:%SZ')


def generate_unicode_uuid(dashed=True):
    return uuidutils.generate_uuid(dashed=dashed)


def validate_int_in_range(name, value, min_allowed, max_allowed):
    unit_mapping = {
        "cpu": "millicpu",
        "memory": "bytes",
        "timeout": "seconds"
    }

    try:
        value_int = int(value)
    except ValueError:
        raise exc.InputException(
            'Invalid %s resource specified. An integer is required.' % name
        )

    if (value_int < min_allowed or value_int > max_allowed):
        raise exc.InputException(
            '%s resource limitation not within the allowable range: '
            '%s ~ %s(%s).' %
            (name, min_allowed, max_allowed, unit_mapping[name])
        )


def disable_ssl_warnings(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="A true SSLContext object is not available"
            )
            warnings.filterwarnings(
                "ignore",
                message="Unverified HTTPS request is being made"
            )
            return func(*args, **kwargs)

    return wrapper


class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used from a forked multiprocessing child.

    Usage:
    from qinling.utils import common
    common.ForkedPdb().set_trace()
    """

    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin', 'r')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin


def md5(file=None, content=None):
    hash_md5 = hashlib.md5()

    if file:
        with open(file, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
    elif content:
        hash_md5.update(content)

    return hash_md5.hexdigest()



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import abc

import six
from stevedore import driver

from qinling import exceptions as exc

ORCHESTRATOR = None


@six.add_metaclass(abc.ABCMeta)
class OrchestratorBase(object):
    """OrchestratorBase interface."""

    @abc.abstractmethod
    def create_pool(self, name, image, trusted=True, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def delete_pool(self, name, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def update_pool(self, name, image=None, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def get_pool(self, name, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_execution(self, function_id, function_version, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def run_execution(self, execution_id, function_id, function_version,
                      **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def delete_function(self, function_id, function_version, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def scaleup_function(self, function_id, function_version, **kwargs):
        raise NotImplementedError

    @abc.abstractmethod
    def delete_worker(self, worker_name, **kwargs):
        raise NotImplementedError


def load_orchestrator(conf, qinling_endpoint):
    global ORCHESTRATOR

    if not ORCHESTRATOR:
        try:
            mgr = driver.DriverManager('qinling.orchestrator',
                                       conf.engine.orchestrator,
                                       invoke_on_load=True,
                                       invoke_args=[conf, qinling_endpoint])

            ORCHESTRATOR = mgr.driver
        except Exception as e:
            raise exc.OrchestratorException(
                'Failed to load orchestrator: %s. Error: %s' %
                (conf.engine.orchestrator, str(e))
            )

    return ORCHESTRATOR



# Copyright 2018 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tempest.lib import decorators
from tempest.lib import exceptions
import tenacity

from qinling_tempest_plugin.tests import base


class FunctionVersionsTest(base.BaseQinlingTest):
    name_prefix = 'FunctionVersionsTest'

    def setUp(self):
        super(FunctionVersionsTest, self).setUp()

        # Wait until runtime is available
        self.wait_runtime_available(self.runtime_id)

    @decorators.idempotent_id('ce630c59-a79d-4b2d-89af-c7c5c8f8bd3f')
    def test_create(self):
        function_id = self.create_function()
        new_version = self.create_function_version(function_id)

        self.assertEqual(1, new_version)

        resp, body = self.client.get_resources(
            'functions/%s/versions' % function_id)

        self.assertEqual(200, resp.status)
        self.assertIn(
            new_version,
            [v['version_number'] for v in body['function_versions']]
        )

    @decorators.idempotent_id('9da2d24c-2ce4-4e6f-9e44-74ef1b9ec3cc')
    def test_create_function_no_change(self):
        function_id = self.create_function()
        self.create_function_version(function_id)

        self.assertRaises(
            exceptions.Forbidden,
            self.client.create_function_version,
            function_id
        )

    @decorators.idempotent_id('6864d134-fbb9-4738-9721-b541c4362789')
    def test_create_function_change(self):
        function_id = self.create_function()
        version_1 = self.create_function_version(function_id)
        self.update_function_package(function_id,
                                     "python/test_python_sleep.py")
        version_2 = self.create_function_version(function_id)

        self.assertGreater(version_2, version_1)

        resp, body = self.client.get_resources(
            'functions/%s/versions' % function_id)
        self.assertEqual(200, resp.status)

        numbers = [v['version_number'] for v in body['function_versions']]
        self.assertIn(version_1, numbers)
        self.assertIn(version_2, numbers)

    @decorators.idempotent_id('3f735ed4-64b0-4ec3-8bf2-507e38dcea19')
    def test_create_admin_not_allowed(self):
        """test_create_admin_not_allowed

        Even admin user can not create function version for normal user's
        function.
        """
        function_id = self.create_function()

        self.assertRaises(
            exceptions.NotFound,
            self.admin_client.create_function_version,
            function_id
        )

    # @decorators.idempotent_id('78dc5552-fcb8-4b27-86f7-5f3d96143934')
    # def test_create_version_lock_failed(self):
    #     """test_create_version_lock_failed
    #
    #     Creating a function requires a lock. If qinling failed to acquire the
    #     lock then an error would be returned after some retries.
    #
    #     In this test we acquire the lock manually, so that qinling will fail
    #     to acquire the lock.
    #     """
    #     function_id = self.create_function()
    #
    #     from qinling_tempest_plugin.tests import utils
    #     etcd3_client = utils.get_etcd_client()
    #     lock_id = "function_version_%s" % function_id
    #     with etcd3_client.lock(id=lock_id):
    #         self.assertRaises(
    #             exceptions.ServerFault,
    #             self.client.create_function_version,
    #             function_id
    #         )

    @decorators.idempotent_id('43c06f41-d116-43a7-a61c-115f7591b22e')
    def test_get_by_admin(self):
        """Admin user can get normal user's function version."""
        function_id = self.create_function()
        version = self.create_function_version(function_id)

        resp, body = self.admin_client.get_function_version(function_id,
                                                            version)

        self.assertEqual(200, resp.status)
        self.assertEqual(version, body.get("version_number"))

    @decorators.idempotent_id('e6b865d8-ffa8-4cfc-8afb-820c64f9b2af')
    def test_get_all_by_admin(self):
        """Admin user can list normal user's function version."""
        function_id = self.create_function()
        version = self.create_function_version(function_id)

        resp, body = self.admin_client.get_function_versions(function_id)

        self.assertEqual(200, resp.status)
        self.assertIn(
            version,
            [v['version_number'] for v in body['function_versions']]
        )

    @decorators.idempotent_id('0e70ef18-687c-4ce4-ae29-aee2f88b4b9c')
    def test_delete(self):
        function_id = self.create_function()
        version = self.create_function_version(function_id)

        resp = self.client.delete_function_version(function_id, version)

        self.assertEqual(204, resp.status)

        resp, body = self.client.get_function_versions(function_id)

        self.assertEqual(200, resp.status)
        self.assertNotIn(
            version,
            [v['version_number'] for v in body['function_versions']]
        )

    @decorators.idempotent_id('c6717e2e-e80a-43d9-a25b-84f4b7453c76')
    def test_delete_by_admin(self):
        """test_delete_by_admin

        Admin user can not delete normal user's function version.
        """
        function_id = self.create_function()
        version = self.create_function_version(function_id)

        self.assertRaises(
            exceptions.NotFound,
            self.admin_client.delete_function_version,
            function_id,
            version
        )

    @decorators.idempotent_id('7898f89f-a490-42a3-8cf7-63cbd9543a06')
    def test_detach(self):
        """Admin only operation."""
        function_id = self.create_function()
        version = self.create_function_version(function_id)

        # Create execution to allocate worker
        resp, _ = self.client.create_execution(
            function_id, input='{"name": "Qinling"}', version=version
        )
        self.assertEqual(201, resp.status)

        resp, body = self.admin_client.get_function_workers(function_id,
                                                            version=version)
        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['workers']))

        # Detach function version from workers
        resp, _ = self.admin_client.detach_function(function_id,
                                                    version=version)
        self.assertEqual(202, resp.status)

        def _assert_workers():
            resp, body = self.admin_client.get_function_workers(
                function_id,
                version=version
            )
            self.assertEqual(200, resp.status)
            self.assertEqual(0, len(body['workers']))

        r = tenacity.Retrying(
            wait=tenacity.wait_fixed(1),
            stop=tenacity.stop_after_attempt(5),
            retry=tenacity.retry_if_exception_type(AssertionError)
        )
        r.call(_assert_workers)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from keystoneauth1.identity import v3
from keystoneauth1 import session
from keystoneclient.v3 import client as ks_client
from oslo_config import cfg
from oslo_log import log as logging
import swiftclient

from qinling import context
from qinling.utils import common

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


def _get_user_keystone_session():
    ctx = context.get_ctx()

    auth = v3.Token(
        auth_url=CONF.keystone_authtoken.www_authenticate_uri,
        token=ctx.auth_token,
        project_domain_name=ctx.project_domain_name,
        project_name=ctx.project_name
    )

    return session.Session(auth=auth, verify=False)


@common.disable_ssl_warnings
def get_swiftclient():
    session = _get_user_keystone_session()

    conn = swiftclient.Connection(session=session)

    return conn


@common.disable_ssl_warnings
def get_user_client():
    ctx = context.get_ctx()
    auth_url = CONF.keystone_authtoken.www_authenticate_uri
    client = ks_client.Client(
        user_id=ctx.user,
        token=ctx.auth_token,
        tenant_id=ctx.projectid,
        auth_url=auth_url
    )
    client.management_url = auth_url

    return client


@common.disable_ssl_warnings
def get_service_client():
    client = ks_client.Client(
        username=CONF.keystone_authtoken.username,
        password=CONF.keystone_authtoken.password,
        project_name=CONF.keystone_authtoken.project_name,
        auth_url=CONF.keystone_authtoken.www_authenticate_uri,
        user_domain_name=CONF.keystone_authtoken.user_domain_name,
        project_domain_name=CONF.keystone_authtoken.project_domain_name
    )
    return client


@common.disable_ssl_warnings
def get_trust_client(trust_id):
    """Get project keystone client using admin credential."""
    client = ks_client.Client(
        username=CONF.keystone_authtoken.username,
        password=CONF.keystone_authtoken.password,
        auth_url=CONF.keystone_authtoken.www_authenticate_uri,
        trust_id=trust_id
    )

    return client


@common.disable_ssl_warnings
def create_trust():
    ctx = context.get_ctx()
    user_client = get_user_client()
    trustee_id = get_service_client().user_id

    return user_client.trusts.create(
        trustor_user=ctx.user,
        trustee_user=trustee_id,
        impersonation=True,
        role_names=ctx.roles,
        project=ctx.tenant
    )


@common.disable_ssl_warnings
def delete_trust(trust_id):
    """Delete trust from keystone.

    The trust can only be deleted by original user(trustor)
    """
    if not trust_id:
        return

    try:
        client = get_user_client()
        client.trusts.delete(trust_id)
        LOG.debug('Trust %s deleted.', trust_id)
    except Exception:
        LOG.exception("Failed to delete trust [id=%s]", trust_id)


def create_trust_context(trust_id, project_id):
    """Creates Qinling context on behalf of the project."""
    if CONF.pecan.auth_enable:
        client = get_trust_client(trust_id)

        return context.Context(
            user=client.user_id,
            tenant=project_id,
            auth_token=client.auth_token,
            is_trust_scoped=True,
            trust_id=trust_id,
        )

    return context.Context(
        user=None,
        tenant=context.DEFAULT_PROJECT_ID,
        auth_token=None,
        is_admin=True
    )


def get_qinling_endpoint():
    '''Get Qinling service endpoint.'''
    if CONF.qinling_endpoint:
        return CONF.qinling_endpoint

    region = CONF.keystone_authtoken.region_name
    auth = v3.Password(
        auth_url=CONF.keystone_authtoken.www_authenticate_uri,
        username=CONF.keystone_authtoken.username,
        password=CONF.keystone_authtoken.password,
        project_name=CONF.keystone_authtoken.project_name,
        user_domain_name=CONF.keystone_authtoken.user_domain_name,
        project_domain_name=CONF.keystone_authtoken.project_domain_name,
    )
    sess = session.Session(auth=auth, verify=False)
    endpoint = sess.get_endpoint(service_type='function-engine',
                                 interface='public',
                                 region_name=region)

    return endpoint



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
import sys
import zipfile

from flask import Flask
from flask import make_response
from flask import request
from oslo_concurrency import lockutils
import requests

app = Flask(__name__)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
ch.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
del app.logger.handlers[:]
app.logger.addHandler(ch)

DOWNLOAD_ERROR = "Failed to download function package from %s, error: %s"


def log(message, level="info"):
    global app
    log_func = getattr(app.logger, level)
    log_func(message)


@lockutils.synchronized('download_function', external=True,
                        lock_path='/var/lock/qinling')
def _download_package(url, zip_file, token=None, unzip=None):
    """Download package and unzip as needed.

    Return None if successful otherwise a Flask.Response object.
    """
    if os.path.isfile(zip_file):
        return None

    log("Start downloading function")

    headers = {}
    if token:
        headers = {'X-Auth-Token': token}

    try:
        r = requests.get(url, headers=headers, stream=True, timeout=30,
                         verify=False)
        if r.status_code != 200:
            return make_response(DOWNLOAD_ERROR % (url, r.content), 500)

        with open(zip_file, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=65535):
                fd.write(chunk)

        log("Downloaded function package to %s" % zip_file)

        if unzip:
            dest = zip_file.split('.')[0]
            with open(zip_file, 'rb') as f:
                zf = zipfile.ZipFile(f)
                zf.extractall(dest)
            log("Unzipped")
    except Exception as e:
        return make_response(DOWNLOAD_ERROR % (url, str(e)), 500)


@app.route('/download', methods=['POST'])
def download():
    """Download function package to a shared folder.

    The parameters 'download_url' and 'function_id' need to be specified
    explicitly. It's guaranteed on the server side.

    :param download_url: The URL for function package download. It's a Qinling
        function resource URL with 'download' enabled.
    :param function_id: Function ID.
    :param token: Optional. The token used for download.
    :param unzip: Optional. If unzip is needed after download.
    """
    params = request.get_json()
    zip_file = '/var/qinling/packages/%s.zip' % params['function_id']
    log("Function package download request received, params: %s" % params)

    resp = _download_package(
        params['download_url'],
        zip_file,
        token=params.get('token'),
        unzip=params.get('unzip', True)
    )

    return resp if resp else 'downloaded'



# Copyright 2018 OpenStack Foundation.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from oslo_config import cfg
from oslo_log import log as logging
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

POST_REQUIRED = set(['name', 'function_id'])
UPDATE_ALLOWED = set(['function_id', 'function_version', 'description'])


class FunctionAliasesController(rest.RestController):
    def __init__(self, *args, **kwargs):
        self.type = 'function_alias'

        super(FunctionAliasesController, self).__init__(*args, **kwargs)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.FunctionAlias,
        body=resources.FunctionAlias,
        status_code=201
    )
    def post(self, body):
        """Create a new alias for the specified function.

        The supported body params:
            - function_id: Required. Function id the alias points to.
            - name: Required. Alias name, must be unique within the project.
            - function_version: Optional. Version number the alias points to.
            - description: Optional. The description of the new alias.
        """
        ctx = context.get_ctx()
        acl.enforce('function_alias:create', ctx)

        params = body.to_dict()
        if not POST_REQUIRED.issubset(set(params.keys())):
            raise exc.InputException(
                'Required param is missing. Required: %s' % POST_REQUIRED
            )
        LOG.info("Creating Alias, params: %s", params)

        values = {
            'function_id': params.get('function_id'),
            'name': params.get('name'),
            'function_version': params.get('function_version'),
            'description': params.get('description'),
        }

        alias = db_api.create_function_alias(**values)

        LOG.info("New alias created.")
        return resources.FunctionAlias.from_db_obj(alias)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.FunctionAlias, wtypes.text)
    def get(self, alias_name):
        acl.enforce('function_alias:get', context.get_ctx())
        LOG.info("Getting function alias  %s.", alias_name)

        alias = db_api.get_function_alias(alias_name)

        return resources.FunctionAlias.from_db_obj(alias)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.FunctionAliases, bool, wtypes.text)
    def get_all(self, all_projects=False, project_id=None):
        """Get all the function aliases.

        :param project_id: Optional. Admin user can query other projects
            resources, the param is ignored for normal user.
        :param all_projects: Optional. Get resources of all projects.
        """
        ctx = context.get_ctx()
        project_id, all_projects = rest_utils.get_project_params(
            project_id, all_projects
        )
        if all_projects:
            acl.enforce('function_version:get_all:all_projects', ctx)

        filters = rest_utils.get_filters(project_id=project_id)

        LOG.info("Get all function aliases. filters=%s", filters)

        db_aliases = db_api.get_function_aliases(
            insecure=all_projects, **filters)
        aliases = [resources.FunctionAlias.from_db_obj(db_model)
                   for db_model in db_aliases]

        return resources.FunctionAliases(function_aliases=aliases)

    @wsme_pecan.wsexpose(None, wtypes.text, status_code=204)
    def delete(self, alias_name):
        """Delete a specific alias.

        """
        ctx = context.get_ctx()
        acl.enforce('function_alias:delete', ctx)
        LOG.info("Deleting alias %s.", alias_name)

        db_api.delete_function_alias(alias_name)

        LOG.info("Alias %s deleted.", alias_name)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.FunctionAlias,
        wtypes.text,
        body=resources.FunctionAlias,
    )
    def put(self, alias_name, body):
        """Update alias for the specified function.

        The supported body params:
            - function_id: Optional. Function id the alias point to.
            - function_version: Optional. Version number the alias point to.
            - description: Optional. The description of the alias.
        """
        ctx = context.get_ctx()
        acl.enforce('function_alias:update', ctx)

        params = body.to_dict()
        values = {}
        for key in UPDATE_ALLOWED:
            if params.get(key) is not None:
                values.update({key: params[key]})
        LOG.info("Updating Alias %s, params: %s", alias_name, values)

        alias = db_api.update_function_alias(alias_name, **values)

        LOG.info("Alias %s updated.", alias_name)
        return resources.FunctionAlias.from_db_obj(alias)



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import resource


def main(number=1024, **kwargs):
    for name, desc in [
        ('RLIMIT_NOFILE', 'number of open files'),
    ]:
        limit_num = getattr(resource, name)
        soft, hard = resource.getrlimit(limit_num)
        print('Maximum %-25s (%-15s) : %20s %20s' % (desc, name, soft, hard))

    files = []

    try:
        for i in range(0, number):
            files.append(_create_file(i))
    finally:
        for f in files:
            f.close()


def _create_file(index):
    f = open('file_%s' % index, 'w')
    return f



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from oslo_utils import uuidutils
from qinling.tests.unit.api import base


class TestFunctionWorkerController(base.APITest):
    @mock.patch('qinling.utils.etcd_util.get_workers')
    def test_get_all_workers(self, mock_get_workers):
        function_id = uuidutils.generate_uuid()
        mock_get_workers.return_value = ['test_worker0', 'test_worker1']

        resp = self.app.get('/v1/functions/%s/workers' % function_id)
        self.assertEqual(200, resp.status_int)
        self._assert_multiple_items(
            resp.json['workers'], 2, function_id=function_id
        )
        self._assert_single_item(
            resp.json['workers'], worker_name='test_worker0'
        )
        self._assert_single_item(
            resp.json['workers'], worker_name='test_worker1'
        )

    @mock.patch('qinling.utils.etcd_util.get_workers')
    def test_get_all_version_workers(self, mock_get_workers):
        function_id = uuidutils.generate_uuid()
        mock_get_workers.return_value = ['test_worker0', 'test_worker1']

        resp = self.app.get(
            '/v1/functions/%s/workers?function_version=1' % function_id
        )

        self.assertEqual(200, resp.status_int)
        mock_get_workers.assert_called_once_with(function_id, version=1)
        self._assert_multiple_items(
            resp.json['workers'],
            2,
            function_id=function_id,
            function_version=1
        )
        self._assert_single_item(
            resp.json['workers'], worker_name='test_worker0'
        )
        self._assert_single_item(
            resp.json['workers'], worker_name='test_worker1'
        )

    def test_get_all_version_workers_not_int(self):
        function_id = uuidutils.generate_uuid()
        resp = self.app.get(
            '/v1/functions/%s/workers?function_version=invalid' % function_id,
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json

import wsme
from wsme import types as wtypes

from qinling.api.controllers.v1 import types


class Resource(wtypes.Base):
    """REST API Resource."""

    _wsme_attributes = []

    def to_dict(self):
        d = {}

        for attr in self._wsme_attributes:
            attr_val = getattr(self, attr.name)
            if not isinstance(attr_val, wtypes.UnsetType):
                d[attr.name] = attr_val

        return d

    @classmethod
    def from_dict(cls, d):
        obj = cls()

        for key, val in d.items():
            if hasattr(obj, key):
                setattr(obj, key, val)

        return obj

    @classmethod
    def from_db_obj(cls, db_obj):
        return cls.from_dict(db_obj.to_dict())

    def __str__(self):
        """WSME based implementation of __str__."""

        res = "%s [" % type(self).__name__

        first = True
        for attr in self._wsme_attributes:
            if not first:
                res += ', '
            else:
                first = False

            res += "%s='%s'" % (attr.name, getattr(self, attr.name))

        return res + "]"

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def get_fields(cls):
        obj = cls()

        return [attr.name for attr in obj._wsme_attributes]


class ResourceList(Resource):
    """Resource containing the list of other resources."""

    next = wtypes.text
    """A link to retrieve the next subset of the resource list"""

    @property
    def collection(self):
        return getattr(self, self._type)

    @classmethod
    def convert_with_links(cls, resources, limit, url=None, fields=None,
                           **kwargs):
        resource_collection = cls()

        setattr(resource_collection, resource_collection._type, resources)

        resource_collection.next = resource_collection.get_next(
            limit,
            url=url,
            fields=fields,
            **kwargs
        )

        return resource_collection

    def has_next(self, limit):
        """Return whether resources has more items."""
        return len(self.collection) and len(self.collection) == limit

    def get_next(self, limit, url=None, fields=None, **kwargs):
        """Return a link to the next subset of the resources."""
        if not self.has_next(limit):
            return wtypes.Unset

        q_args = ''.join(
            ['%s=%s&' % (key, value) for key, value in kwargs.items()]
        )

        resource_args = (
            '?%(args)slimit=%(limit)d&marker=%(marker)s' %
            {
                'args': q_args,
                'limit': limit,
                'marker': self.collection[-1].id
            }
        )

        # Fields is handled specially here, we can move it above when it's
        # supported by all resources query.
        if fields:
            resource_args += '&fields=%s' % fields

        next_link = "%(host_url)s/v2/%(resource)s%(args)s" % {
            'host_url': url,
            'resource': self._type,
            'args': resource_args
        }

        return next_link

    def to_dict(self):
        d = {}

        for attr in self._wsme_attributes:
            attr_val = getattr(self, attr.name)

            if isinstance(attr_val, list):
                if isinstance(attr_val[0], Resource):
                    d[attr.name] = [v.to_dict() for v in attr_val]
            elif not isinstance(attr_val, wtypes.UnsetType):
                d[attr.name] = attr_val

        return d


class Link(Resource):
    """Web link."""

    href = wtypes.text
    target = wtypes.text
    rel = wtypes.text


class Function(Resource):
    id = wtypes.text
    name = wtypes.text
    description = wtypes.text
    cpu = int
    memory_size = int
    timeout = int
    runtime_id = wsme.wsattr(types.uuid, readonly=True)
    code = types.jsontype
    entry = wtypes.text
    count = wsme.wsattr(int, readonly=True)
    latest_version = wsme.wsattr(int, readonly=True)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wtypes.text
    updated_at = wtypes.text


class Functions(ResourceList):
    functions = [Function]

    def __init__(self, **kwargs):
        self._type = 'functions'

        super(Functions, self).__init__(**kwargs)


class FunctionWorker(Resource):
    function_id = wsme.wsattr(types.uuid, readonly=True)
    function_version = wsme.wsattr(int, readonly=True)
    worker_name = wsme.wsattr(wtypes.text, readonly=True)


class FunctionWorkers(ResourceList):
    workers = [FunctionWorker]

    def __init__(self, **kwargs):
        self._type = 'workers'
        super(FunctionWorkers, self).__init__(**kwargs)


class Runtime(Resource):
    id = wtypes.text
    name = wtypes.text
    image = wtypes.text
    description = wtypes.text
    is_public = wsme.wsattr(bool, default=True)
    trusted = bool
    status = wsme.wsattr(wtypes.text, readonly=True)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)


class Runtimes(ResourceList):
    runtimes = [Runtime]

    def __init__(self, **kwargs):
        self._type = 'environments'

        super(Runtimes, self).__init__(**kwargs)


class RuntimePoolCapacity(Resource):
    total = wsme.wsattr(int, readonly=True)
    available = wsme.wsattr(int, readonly=True)


class RuntimePool(Resource):
    name = wsme.wsattr(wtypes.text, readonly=True)
    capacity = wsme.wsattr(RuntimePoolCapacity, readonly=True)


class Execution(Resource):
    id = types.uuid
    function_id = wsme.wsattr(types.uuid)
    function_version = wsme.wsattr(int, default=0)
    function_alias = wsme.wsattr(wtypes.text)
    description = wtypes.text
    status = wsme.wsattr(wtypes.text, readonly=True)
    sync = bool
    input = wtypes.text
    result = wsme.wsattr(types.jsontype, readonly=True)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)

    @classmethod
    def from_dict(cls, d):
        obj = cls()

        for key, val in d.items():
            if key == 'input' and val is not None:
                if val.get('__function_input'):
                    setattr(obj, key, val.get('__function_input'))
                else:
                    setattr(obj, key, json.dumps(val))
                continue
            if hasattr(obj, key):
                setattr(obj, key, val)

        return obj


class Executions(ResourceList):
    executions = [Execution]

    def __init__(self, **kwargs):
        self._type = 'executions'

        super(Executions, self).__init__(**kwargs)


class Job(Resource):
    id = types.uuid
    name = wtypes.text
    function_id = types.uuid
    function_alias = wtypes.text
    function_version = wsme.wsattr(int, default=0)
    function_input = wtypes.text
    status = wtypes.text
    pattern = wtypes.text
    count = int
    first_execution_time = wtypes.text
    next_execution_time = wtypes.text
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)


class Jobs(ResourceList):
    jobs = [Job]

    def __init__(self, **kwargs):
        self._type = 'jobs'

        super(Jobs, self).__init__(**kwargs)


class ScaleInfo(Resource):
    count = wtypes.IntegerType(minimum=1)


class Webhook(Resource):
    id = types.uuid
    function_id = types.uuid
    function_alias = wtypes.text
    function_version = wsme.wsattr(int)
    description = wtypes.text
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)
    webhook_url = wsme.wsattr(wtypes.text, readonly=True)


class Webhooks(ResourceList):
    webhooks = [Webhook]

    def __init__(self, **kwargs):
        self._type = 'webhooks'

        super(Webhooks, self).__init__(**kwargs)


class FunctionVersion(Resource):
    id = types.uuid
    description = wtypes.text
    function_id = wsme.wsattr(types.uuid, readonly=True)
    version_number = wsme.wsattr(int, readonly=True)
    count = wsme.wsattr(int, readonly=True)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)


class FunctionVersions(ResourceList):
    function_versions = [FunctionVersion]

    def __init__(self, **kwargs):
        self._type = 'function_versions'

        super(FunctionVersions, self).__init__(**kwargs)


class FunctionAlias(Resource):
    id = types.uuid
    name = wtypes.text
    description = wtypes.text
    function_id = types.uuid
    function_version = wsme.wsattr(int)
    project_id = wsme.wsattr(wtypes.text, readonly=True)
    created_at = wsme.wsattr(wtypes.text, readonly=True)
    updated_at = wsme.wsattr(wtypes.text, readonly=True)


class FunctionAliases(ResourceList):
    function_aliases = [FunctionAlias]

    def __init__(self, **kwargs):
        self._type = 'function_aliases'

        super(FunctionAliases, self).__init__(**kwargs)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools

from oslo_config import cfg
from oslo_db import options as db_options
from oslo_db.sqlalchemy import session as db_session

from qinling import context
from qinling import exceptions as exc
from qinling.utils import thread_local

# Note(dzimine): sqlite only works for basic testing.
db_options.set_defaults(cfg.CONF, connection="sqlite:///qinling.sqlite")
_FACADE = None
_DB_SESSION_THREAD_LOCAL_NAME = "db_sql_alchemy_session"


def _get_facade():
    global _FACADE
    if _FACADE is None:
        _FACADE = db_session.EngineFacade.from_config(cfg.CONF, sqlite_fk=True)
    return _FACADE


def get_session(expire_on_commit=False, autocommit=False):
    """Helper method to grab session."""
    facade = _get_facade()
    return facade.get_session(expire_on_commit=expire_on_commit,
                              autocommit=autocommit)


def get_engine():
    facade = _get_facade()
    return facade.get_engine()


def _get_thread_local_session():
    return thread_local.get_thread_local(_DB_SESSION_THREAD_LOCAL_NAME)


def _get_or_create_thread_local_session():
    ses = _get_thread_local_session()

    if ses:
        return ses, False

    ses = get_session()
    _set_thread_local_session(ses)

    return ses, True


def _set_thread_local_session(session):
    thread_local.set_thread_local(_DB_SESSION_THREAD_LOCAL_NAME, session)


def start_tx():
    """Starts transaction.

    Opens new database session and starts new transaction assuming
    there wasn't any opened sessions within the same thread.
    """
    if _get_thread_local_session():
        raise exc.DBError(
            "Database transaction has already been started."
        )

    _set_thread_local_session(get_session())


def commit_tx():
    """Commits previously started database transaction."""
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DBError(
            "Nothing to commit. Database transaction"
            " has not been previously started."
        )

    ses.commit()


def rollback_tx():
    """Rolls back previously started database transaction."""
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DBError(
            "Nothing to roll back. Database transaction has not been started."
        )

    ses.rollback()


def end_tx():
    """Ends transaction.

    Ends current database transaction.
    It rolls back all uncommitted changes and closes database session.
    """
    ses = _get_thread_local_session()

    if not ses:
        raise exc.DBError(
            "Database transaction has not been started."
        )

    if ses.dirty:
        rollback_tx()

    ses.close()
    _set_thread_local_session(None)


def session_aware():
    """Decorator for methods working within db session."""

    def _decorator(func):
        @functools.wraps(func)
        def _within_session(*args, **kw):
            ses, created = _get_or_create_thread_local_session()

            try:
                kw['session'] = ses

                result = func(*args, **kw)

                if created:
                    ses.commit()

                return result
            except Exception:
                if created:
                    ses.rollback()
                raise
            finally:
                if created:
                    _set_thread_local_session(None)
                    ses.close()

        return _within_session

    return _decorator


def insecure_aware():
    """Decorator for methods working within insecure db query or not."""

    def _decorator(func):
        @functools.wraps(func)
        def _with_insecure(*args, **kw):
            if kw.get('insecure') is None:
                insecure = context.get_ctx().is_admin
                kw['insecure'] = insecure
            return func(*args, **kw)

        return _with_insecure

    return _decorator


@session_aware()
def model_query(model, columns=(), session=None):
    """Query helper.

    :param model: Base model to query.
    :param columns: Optional. Which columns to be queried.
    """
    if columns:
        return session.query(*columns)

    return session.query(model)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import time

from oslo_log import log as logging
import requests
import six
import tenacity

from qinling import context
from qinling.db import api as db_api
from qinling import status
from qinling.utils import constants

LOG = logging.getLogger(__name__)


def url_request(request_session, url, body=None):
    """Send request to a service url."""
    exception = None

    # Send ping request first to make sure the url works
    try:
        temp = url.split('/')
        temp[-1] = 'ping'
        ping_url = '/'.join(temp)
        r = tenacity.Retrying(
            wait=tenacity.wait_fixed(1),
            stop=tenacity.stop_after_attempt(30),
            reraise=True,
            retry=tenacity.retry_if_exception_type(IOError)
        )
        r.call(request_session.get, ping_url, timeout=(3, 3), verify=False)
    except Exception as e:
        LOG.exception(
            "Failed to request url %s, error: %s", ping_url, str(e)
        )
        return False, {'output': 'Function execution failed.'}

    for a in six.moves.xrange(10):
        res = None
        try:
            # Default execution max duration is 3min, could be configurable
            res = request_session.post(
                url, json=body, timeout=(3, 180), verify=False
            )
            return True, res.json()
        except requests.ConnectionError as e:
            exception = e
            time.sleep(1)
        except Exception as e:
            LOG.exception(
                "Failed to request url %s, error: %s", url, str(e)
            )
            if res:
                LOG.error("Response status: %s, content: %s",
                          res.status_code, res.content)

            return False, {'output': 'Function execution timeout.'}

    LOG.exception("Could not connect to function service. Reason: %s",
                  exception)

    return False, {'output': 'Internal service error.'}


def get_request_data(conf, function_id, version, execution_id, rlimit, input,
                     entry, trust_id, qinling_endpoint, timeout):
    """Prepare the request body should send to the worker."""
    ctx = context.get_ctx()

    if version == 0:
        download_url = (
            '%s/%s/functions/%s?download=true' %
            (qinling_endpoint.strip('/'), constants.CURRENT_VERSION,
             function_id)
        )
    else:
        download_url = (
            '%s/%s/functions/%s/versions/%s?download=true' %
            (qinling_endpoint.strip('/'), constants.CURRENT_VERSION,
             function_id, version)
        )

    data = {
        'execution_id': execution_id,
        'cpu': rlimit['cpu'],
        'memory_size': rlimit['memory_size'],
        'input': input,
        'function_id': function_id,
        'function_version': version,
        'entry': entry,
        'download_url': download_url,
        'request_id': ctx.request_id,
        'timeout': timeout,
    }
    if conf.pecan.auth_enable:
        data.update(
            {
                'token': ctx.auth_token,
                'auth_url': conf.keystone_authtoken.www_authenticate_uri,
                'username': conf.keystone_authtoken.username,
                'password': conf.keystone_authtoken.password,
                'trust_id': trust_id
            }
        )

    return data


def db_set_execution_status(execution_id, execution_status, logs, res):
    db_api.update_execution(
        execution_id,
        {
            'status': execution_status,
            'logs': logs,
            'result': res
        }
    )


def finish_execution(execution_id, success, res, is_image_source=False):
    logs = res.pop('logs', '')
    success = success and res.pop('success', True)

    LOG.debug(
        'Finished execution %s, success: %s', execution_id, success
    )
    db_set_execution_status(
        execution_id, status.SUCCESS if success else status.FAILED,
        logs, res
    )


def handle_execution_exception(execution_id, exc_str):
    # This method should be called from an exception handler
    LOG.exception(
        'Error running execution %s: %s', execution_id, exc_str
    )
    db_set_execution_status(
        execution_id, status.ERROR,
        '',
        {'output': 'Function execution failed.'}
    )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from datetime import datetime
from datetime import timedelta

from dateutil import parser

from qinling import context as auth_context
from qinling.db import api as db_api
from qinling import status
from qinling.tests.unit.api import base


class TestJobController(base.APITest):
    def setUp(self):
        super(TestJobController, self).setUp()

        # Insert a function record in db for each test case. The data will be
        # removed automatically in db clean up.
        db_function = self.create_function()
        self.function_id = db_function.id

    def test_create_with_function(self):
        body = {
            'name': self.rand_name('job', prefix=self.prefix),
            'first_execution_time': str(
                datetime.utcnow() + timedelta(hours=1)),
            'function_id': self.function_id
        }
        resp = self.app.post_json('/v1/jobs', body)

        self.assertEqual(201, resp.status_int)

    def test_create_with_version(self):
        db_api.increase_function_version(self.function_id, 0)

        body = {
            'name': self.rand_name('job', prefix=self.prefix),
            'first_execution_time': str(
                datetime.utcnow() + timedelta(hours=1)),
            'function_id': self.function_id,
            'function_version': 1,
        }
        resp = self.app.post_json('/v1/jobs', body)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(1, resp.json.get('function_version'))

    def test_create_with_alias(self):
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.function_id,
            'name': name
        }
        db_api.create_function_alias(**body)

        job_body = {
            'name': self.rand_name('job', prefix=self.prefix),
            'first_execution_time': str(
                datetime.utcnow() + timedelta(hours=1)),
            'function_alias': name
        }

        resp = self.app.post_json('/v1/jobs', job_body)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(name, resp.json.get('function_alias'))
        self.assertIsNone(resp.json.get('function_id'))

    def test_create_with_invalid_alias(self):
        body = {
            'function_alias': 'fake_alias',
            'first_execution_time': str(
                datetime.utcnow() + timedelta(hours=1)),
        }

        resp = self.app.post_json('/v1/jobs', body, expect_errors=True)

        self.assertEqual(404, resp.status_int)

    def test_create_without_required_params(self):
        resp = self.app.post(
            '/v1/jobs',
            params={},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

    def test_create_pattern(self):
        body = {
            'name': self.rand_name('job', prefix=self.prefix),
            'function_id': self.function_id,
            'pattern': '0 21 * * *',
            'count': 10
        }
        resp = self.app.post_json('/v1/jobs', body)

        self.assertEqual(201, resp.status_int)

        res = resp.json
        self.assertEqual(
            res["first_execution_time"],
            res["next_execution_time"]
        )

    def test_create_both_pattern_and_first_execution_time(self):
        body = {
            'name': self.rand_name('job', prefix=self.prefix),
            'function_id': self.function_id,
            'pattern': '0 21 * * *',
            'first_execution_time': str(
                datetime.utcnow() + timedelta(hours=1)),
            'count': 10
        }
        resp = self.app.post_json('/v1/jobs', body)

        self.assertEqual(201, resp.status_int)

        res = resp.json
        self.assertGreaterEqual(
            parser.parse(res["next_execution_time"], ignoretz=True),
            parser.parse(res["first_execution_time"], ignoretz=True)
        )

    def test_delete(self):
        job_id = self.create_job(
            self.function_id,
            first_execution_time=datetime.utcnow(),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
            status=status.RUNNING,
            count=1
        ).id

        resp = self.app.delete('/v1/jobs/%s' % job_id)

        self.assertEqual(204, resp.status_int)

    def test_update_one_shot_job(self):
        job_id = self.create_job(
            self.function_id,
            first_execution_time=datetime.utcnow(),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
            status=status.RUNNING,
            count=1
        ).id

        req_body = {
            'name': 'new_name',
            'status': status.PAUSED
        }
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

        req_body = {
            'status': status.RUNNING
        }
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

    def test_update_one_shot_job_failed(self):
        job_id = self.create_job(
            self.function_id,
            first_execution_time=datetime.utcnow(),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
            status=status.RUNNING,
            count=1
        ).id
        url = '/v1/jobs/%s' % job_id

        # Try to change job type
        resp = self.app.put_json(
            url,
            {'pattern': '*/1 * * * *'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('Can not change job type.', resp.json['faultstring'])

        # Try to resume job but the execution time is invalid
        auth_context.set_ctx(self.ctx)
        self.addCleanup(auth_context.set_ctx, None)
        db_api.update_job(
            job_id,
            {
                'next_execution_time': datetime.utcnow() - timedelta(hours=1),
                'status': status.PAUSED
            }
        )
        resp = self.app.put_json(
            url,
            {'status': status.RUNNING},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'Execution time must be at least 1 minute in the future',
            resp.json['faultstring']
        )

    def test_update_recurring_job(self):
        job_id = self.create_job(
            self.function_id,
            first_execution_time=datetime.utcnow() + timedelta(hours=1),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
            pattern='0 */1 * * *',
            status=status.RUNNING,
            count=10
        ).id

        next_hour_and_half = datetime.utcnow() + timedelta(hours=1.5)
        next_two_hours = datetime.utcnow() + timedelta(hours=2)

        req_body = {
            'next_execution_time': str(
                next_hour_and_half.strftime('%Y-%m-%dT%H:%M:%SZ')
            ),
            'pattern': '1 */1 * * *'
        }
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

        # Pause the job and resume with a valid next_execution_time
        req_body = {
            'status': status.PAUSED
        }
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

        req_body = {
            'status': status.RUNNING,
            'next_execution_time': str(
                next_two_hours.strftime('%Y-%m-%dT%H:%M:%SZ')
            ),
        }
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

        # Pause the job and resume without specifying next_execution_time
        auth_context.set_ctx(self.ctx)
        self.addCleanup(auth_context.set_ctx, None)
        db_api.update_job(
            job_id,
            {
                'next_execution_time': datetime.utcnow() - timedelta(hours=1),
                'status': status.PAUSED
            }
        )

        req_body = {'status': status.RUNNING}
        resp = self.app.put_json('/v1/jobs/%s' % job_id, req_body)

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, req_body)

    def test_update_recurring_job_failed(self):
        job_id = self.create_job(
            self.function_id,
            first_execution_time=datetime.utcnow() + timedelta(hours=1),
            next_execution_time=datetime.utcnow() + timedelta(hours=1),
            pattern='0 */1 * * *',
            status=status.RUNNING,
            count=10
        ).id
        url = '/v1/jobs/%s' % job_id

        # Try to change job type
        resp = self.app.put_json(
            url,
            {'pattern': ''},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn('Can not change job type.', resp.json['faultstring'])

        # Pause the job and try to resume with an invalid next_execution_time
        auth_context.set_ctx(self.ctx)
        self.addCleanup(auth_context.set_ctx, None)
        db_api.update_job(job_id, {'status': status.PAUSED})
        resp = self.app.put_json(
            url,
            {
                'status': status.RUNNING,
                'next_execution_time': str(
                    datetime.utcnow() - timedelta(hours=1)
                )
            },
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn(
            'Execution time must be at least 1 minute in the future',
            resp.json['faultstring']
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import status
from qinling.tests.unit.api import base


class TestExecutionController(base.APITest):
    def setUp(self):
        super(TestExecutionController, self).setUp()

        db_func = self.create_function()
        self.func_id = db_func.id

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_create_with_function(self, mock_create_execution):
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)

        self.assertEqual(201, resp.status_int)

        resp = self.app.get('/v1/functions/%s' % self.func_id)

        self.assertEqual(1, resp.json.get('count'))

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_create_with_version(self, mock_rpc):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")
        body = {
            'function_id': self.func_id,
            'function_version': 1
        }

        resp = self.app.post_json('/v1/executions', body)
        self.assertEqual(201, resp.status_int)

        resp = self.app.get('/v1/functions/%s' % self.func_id)
        self.assertEqual(0, resp.json.get('count'))

        resp = self.app.get('/v1/functions/%s/versions/1' % self.func_id)
        self.assertEqual(1, resp.json.get('count'))

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_create_with_alias(self, mock_rpc):
        db_api.increase_function_version(self.func_id, 0,
                                         description="version 1")
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'name': name
        }
        db_api.create_function_alias(**body)

        execution_body = {
            'function_alias': name
        }
        resp = self.app.post_json('/v1/executions', execution_body)
        self.assertEqual(201, resp.status_int)
        self.assertEqual(name, resp.json.get('function_alias'))

        resp = self.app.get('/v1/functions/%s' % self.func_id)
        self.assertEqual(0, resp.json.get('count'))

        resp = self.app.get('/v1/functions/%s/versions/1' % self.func_id)
        self.assertEqual(1, resp.json.get('count'))

    def test_create_with_invalid_alias(self):
        body = {
            'function_alias': 'fake_alias',
        }

        resp = self.app.post_json('/v1/executions', body, expect_errors=True)

        self.assertEqual(404, resp.status_int)

    def test_create_without_required_params(self):
        resp = self.app.post(
            '/v1/executions',
            params={},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_create_rpc_error(self, mock_create_execution):
        mock_create_execution.side_effect = exc.QinlingException
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(status.ERROR, resp.json.get('status'))

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_get(self, mock_create_execution):
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)

        self.assertEqual(201, resp.status_int)

        resp = self.app.get('/v1/executions/%s' % resp.json.get('id'))

        self.assertEqual(self.func_id, resp.json.get('function_id'))

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_get_all(self, mock_create_execution):
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)
        exec_id = resp.json.get('id')

        self.assertEqual(201, resp.status_int)

        resp = self.app.get('/v1/executions')

        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(
            resp.json['executions'], id=exec_id
        )
        self._assertDictContainsSubset(actual, body)

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_get_all_filter(self, mock_create_execution):
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)
        exec_id = resp.json.get('id')
        self.assertEqual(201, resp.status_int)

        # Test filtering by 'function_id'
        resp = self.app.get('/v1/executions?function_id=%s' % self.func_id)
        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(
            resp.json['executions'], id=exec_id
        )
        self._assertDictContainsSubset(actual, body)

        # Test filtering by 'status'
        resp = self.app.get(
            '/v1/executions?function_id=%s&status=running' % self.func_id
        )
        self.assertEqual(200, resp.status_int)
        self._assert_single_item(resp.json['executions'], id=exec_id)

    @mock.patch('qinling.rpc.EngineClient.create_execution')
    def test_delete(self, mock_create_execution):
        body = {
            'function_id': self.func_id,
        }
        resp = self.app.post_json('/v1/executions', body)
        exec_id = resp.json.get('id')

        resp = self.app.delete('/v1/executions/%s' % exec_id)

        self.assertEqual(204, resp.status_int)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from pecan import rest
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import rpc
from qinling import status
from qinling.utils import etcd_util
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)

POST_REQUIRED = set(['image'])
UPDATE_ALLOWED = set(['name', 'description', 'image'])


class RuntimesController(rest.RestController):
    _custom_actions = {'pool': ['GET']}

    def __init__(self, *args, **kwargs):
        self.engine_client = rpc.get_engine_client()
        self.type = 'runtime'

        super(RuntimesController, self).__init__(*args, **kwargs)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Runtime, types.uuid)
    def get(self, id):
        LOG.info("Get resource.", resource={'type': self.type, 'id': id})

        runtime_db = db_api.get_runtime(id)

        return resources.Runtime.from_db_obj(runtime_db)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Runtimes)
    def get_all(self):
        LOG.info("Get all %ss.", self.type)

        runtimes = [resources.Runtime.from_db_obj(db_model)
                    for db_model in db_api.get_runtimes()]

        return resources.Runtimes(runtimes=runtimes)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Runtime,
        body=resources.Runtime,
        status_code=201
    )
    def post(self, runtime):
        acl.enforce('runtime:create', context.get_ctx())

        params = runtime.to_dict()
        if 'trusted' not in params:
            params['trusted'] = True

        if not POST_REQUIRED.issubset(set(params.keys())):
            raise exc.InputException(
                'Required param is missing. Required: %s' % POST_REQUIRED
            )

        LOG.info("Creating %s, params: %s", self.type, params)

        params.update({'status': status.CREATING})

        db_model = db_api.create_runtime(params)
        self.engine_client.create_runtime(db_model.id)

        return resources.Runtime.from_db_obj(db_model)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        acl.enforce('runtime:delete', context.get_ctx())

        LOG.info("Delete resource.", resource={'type': self.type, 'id': id})

        with db_api.transaction():
            runtime_db = db_api.get_runtime(id)

            # Runtime can not be deleted if still associate with functions.
            funcs = db_api.get_functions(insecure=True, runtime_id={'eq': id})
            if len(funcs):
                raise exc.NotAllowedException(
                    'Runtime %s is still in use.' % id
                )

            runtime_db.status = status.DELETING

        # Clean related resources asynchronously
        self.engine_client.delete_runtime(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Runtime,
        types.uuid,
        body=resources.Runtime
    )
    def put(self, id, runtime):
        """Update runtime.

        Currently, we support update name, description, image. When
        updating image, send message to engine for asynchronous
        handling.
        """
        acl.enforce('runtime:update', context.get_ctx())

        values = {}
        for key in UPDATE_ALLOWED:
            if runtime.to_dict().get(key) is not None:
                values.update({key: runtime.to_dict()[key]})

        LOG.info('Update resource, params: %s', values,
                 resource={'type': self.type, 'id': id})

        image = values.get('image')

        with db_api.transaction():
            if image is not None:
                pre_runtime = db_api.get_runtime(id)
                if pre_runtime.status != status.AVAILABLE:
                    raise exc.RuntimeNotAvailableException(
                        'Runtime %s is not available.' % id
                    )

                pre_image = pre_runtime.image
                if pre_image != image:
                    # Ensure there is no function running in the runtime.
                    db_funcs = db_api.get_functions(
                        insecure=True, fields=['id'], runtime_id=id
                    )
                    func_ids = [func.id for func in db_funcs]

                    for id in func_ids:
                        if etcd_util.get_service_url(id):
                            raise exc.NotAllowedException(
                                'Runtime %s is still in use by functions.' % id
                            )

                    values['status'] = status.UPGRADING
                    self.engine_client.update_runtime(
                        id,
                        image=image,
                        pre_image=pre_image,
                    )

            runtime_db = db_api.update_runtime(id, values)

        return resources.Runtime.from_db_obj(runtime_db)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.RuntimePool, types.uuid)
    def pool(self, id):
        """Get the pool information for the runtime.

        This operation should be admin only.

        We don't check the runtime existence, because this function
        also helps us to check the underlying pool even after the runtime
        is already deleted.
        """
        acl.enforce('runtime_pool:get_all', context.get_ctx())

        LOG.info("Getting pool information for runtime %s.", id)
        capacity = self.engine_client.get_runtime_pool(id)
        pool_capacity = resources.RuntimePoolCapacity.from_dict(capacity)

        return resources.RuntimePool.from_dict(
            {"name": id, "capacity": pool_capacity}
        )



# Copyright 2018 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import requests
from tempest.lib import decorators

from qinling_tempest_plugin.tests import base


class WebhooksTest(base.BaseQinlingTest):
    name_prefix = 'WebhooksTest'

    def setUp(self):
        super(WebhooksTest, self).setUp()
        self.wait_runtime_available(self.runtime_id)
        self.function_id = self.create_function()

    @decorators.idempotent_id('37DCD022-32D6-48D1-B90C-31D605DBE53B')
    def test_webhook_invoke(self):
        webhook_id, url = self.create_webhook(self.function_id)
        resp = requests.post(url, data={'name': 'qinling'}, verify=False)
        self.assertEqual(202, resp.status_code)
        resp_exec_id = resp.json().get('execution_id')
        self.addCleanup(self.client.delete_resource, 'executions',
                        resp_exec_id, ignore_notfound=True)

        resp, body = self.client.get_resources(
            'executions',
            {'description': 'has:%s' % webhook_id}
        )
        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))
        exec_id = body['executions'][0]['id']
        self.assertEqual(resp_exec_id, exec_id)
        self.wait_execution_success(exec_id)

        resp, body = self.client.get_execution_log(exec_id)
        self.assertEqual(200, resp.status)
        self.assertIn('qinling', body)

    @decorators.idempotent_id('68605edb-1e36-4953-907d-aa6e2352bb85')
    def test_webhook_with_function_version(self):
        version = self.create_function_version(self.function_id)
        webhook_id, url = self.create_webhook(self.function_id,
                                              version=version)
        resp = requests.post(url, data={'name': 'version_test'}, verify=False)

        self.assertEqual(202, resp.status_code)

        resp_exec_id = resp.json().get('execution_id')
        self.addCleanup(self.client.delete_resource, 'executions',
                        resp_exec_id, ignore_notfound=True)

        resp, body = self.client.get_resources(
            'executions',
            {'description': 'has:%s' % webhook_id}
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))
        exec_id = body['executions'][0]['id']
        self.assertEqual(resp_exec_id, exec_id)
        self.wait_execution_success(exec_id)

        resp, body = self.client.get_execution_log(exec_id)
        self.assertEqual(200, resp.status)
        self.assertIn('version_test', body)

    @decorators.idempotent_id('a5b5eed3-82ee-4ab1-b9ca-9898e4da6b5a')
    def test_webhook_with_function_alias(self):
        version = self.create_function_version(self.function_id)
        function_alias = self.create_function_alias(self.function_id, version)
        webhook_id, url = self.create_webhook(function_alias=function_alias)
        resp = requests.post(url, data={'name': 'alias_test'}, verify=False)

        self.assertEqual(202, resp.status_code)

        resp_exec_id = resp.json().get('execution_id')
        self.addCleanup(self.client.delete_resource, 'executions',
                        resp_exec_id, ignore_notfound=True)

        resp, body = self.client.get_resources(
            'executions',
            {'description': 'has:%s' % webhook_id}
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))
        exec_id = body['executions'][0]['id']
        self.assertEqual(resp_exec_id, exec_id)
        self.wait_execution_success(exec_id)

        resp, body = self.client.get_execution_log(exec_id)
        self.assertEqual(200, resp.status)
        self.assertIn('alias_test', body)

    @decorators.idempotent_id('8e6e4f76-f748-11e7-8ec3-00224d6b7bc1')
    def test_get_all_admin(self):
        """Admin user can get webhooks of other projects"""
        webhook_id, _ = self.create_webhook(self.function_id)

        resp, body = self.admin_client.get_resources(
            'webhooks?all_projects=true'
        )
        self.assertEqual(200, resp.status)
        self.assertIn(
            webhook_id,
            [item['id'] for item in body['webhooks']]
        )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from keystoneauth1 import loading
from keystonemiddleware import auth_token
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors

from qinling import version

CONF = cfg.CONF

launch_opt = cfg.ListOpt(
    'server',
    default=['all'],
    help='Specifies which qinling server to start by the launch script.'
)

default_opts = [
    cfg.StrOpt(
        'qinling_endpoint',
        help='Qinling service endpoint.'
    ),
]

API_GROUP = 'api'
api_opts = [
    cfg.StrOpt('host', default='0.0.0.0', help='Qinling API server host.'),
    cfg.PortOpt('port', default=7070, help='Qinling API server port.'),
    cfg.BoolOpt(
        'enable_ssl_api',
        default=False,
        help='Enable the integrated stand-alone API to service requests'
             'via HTTPS instead of HTTP.'
    ),
    cfg.IntOpt(
        'api_workers',
        default=processutils.get_worker_count(),
        help='Number of workers for Qinling API service '
             'default is equal to the number of CPUs available if that can '
             'be determined, else a default worker count of 1 is returned.'
    ),
    cfg.BoolOpt(
        'enable_job_handler',
        default=True,
        help='Enable job handler.'
    ),
]

PECAN_GROUP = 'pecan'
pecan_opts = [
    cfg.StrOpt(
        'root',
        default='qinling.api.controllers.root.RootController',
        help='Pecan root controller'
    ),
    cfg.ListOpt(
        'modules',
        default=["qinling.api"],
        help='A list of modules where pecan will search for applications.'
    ),
    cfg.BoolOpt(
        'debug',
        default=False,
        help='Enables the ability to display tracebacks in the browser and'
             ' interactively debug during development.'
    ),
    cfg.BoolOpt(
        'auth_enable',
        default=True,
        help='Enables user authentication in pecan.'
    )
]

ENGINE_GROUP = 'engine'
engine_opts = [
    cfg.StrOpt(
        'host',
        default='0.0.0.0',
        help='Name of the engine node. This can be an opaque '
             'identifier. It is not necessarily a hostname, '
             'FQDN, or IP address.'
    ),
    cfg.StrOpt(
        'topic',
        default='qinling_engine',
        help='The message topic that the engine listens on.'
    ),
    cfg.StrOpt(
        'orchestrator',
        default='kubernetes',
        choices=['kubernetes', 'swarm'],
        help='The container orchestrator.'
    ),
    cfg.IntOpt(
        'function_service_expiration',
        default=3600,
        help='Maximum service time in seconds for function in orchestrator.'
    ),
    cfg.IntOpt(
        'function_concurrency',
        default=3,
        help='Maximum number of concurrent executions per function.'
    ),
    cfg.StrOpt(
        'sidecar_image',
        default='openstackqinling/sidecar:0.0.2',
        help='The sidecar image being used together with the worker.'
    ),
]

STORAGE_GROUP = 'storage'
storage_opts = [
    cfg.StrOpt(
        'file_system_dir',
        help='Directory to store function packages.'
    ),
    cfg.StrOpt(
        'provider',
        default='local',
        choices=['local', 'swift'],
        help='Storage provider for function code package.'
    ),
]

KUBERNETES_GROUP = 'kubernetes'
kubernetes_opts = [
    cfg.StrOpt(
        'namespace',
        default='qinling',
        help='Resources scope created by Qinling.'
    ),
    cfg.IntOpt(
        'replicas',
        default=3,
        help='Number of desired replicas in deployment.'
    ),
    cfg.StrOpt(
        'kube_host',
        default='http://127.0.0.1:8001',
        help='Kubernetes server address, e.g. you can start a proxy to the '
             'Kubernetes API server by using "kubectl proxy" command.'
    ),
    cfg.BoolOpt(
        'use_api_certificate',
        default=True,
        help='Whether to use client certificates to connect to the '
             'Kubernetes API server.'
    ),
    cfg.StrOpt(
        'ssl_ca_cert',
        default='/etc/qinling/pki/kubernetes/ca.crt',
        help='Path to the CA certificate for qinling to use to connect to '
             'the Kubernetes API server.'
    ),
    cfg.StrOpt(
        'cert_file',
        default='/etc/qinling/pki/kubernetes/qinling.crt',
        help='Path to the client certificate for qinling to use to '
             'connect to the Kubernetes API server.'
    ),
    cfg.StrOpt(
        'key_file',
        default='/etc/qinling/pki/kubernetes/qinling.key',
        help='Path to the client certificate key file for qinling to use to '
             'connect to the Kubernetes API server.'
    ),
    cfg.StrOpt(
        'log_devel',
        default='INFO',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR'],
        help='Log level for kubernetes operations.'
    ),
    cfg.ListOpt(
        'trusted_cidrs',
        deprecated_for_removal=True,
        item_type=cfg.types.String(),
        default=[],
        help='List of CIDR that have access to the services in '
             'Kubernetes, e.g. trusted_cidrs=127.0.0.1/32,198.72.124.109/32. '
             'If it is empty list, the default value is the host IP address '
             'that the qinling-engine service is running on.'
    )
]

ETCD_GROUP = 'etcd'
etcd_opts = [
    cfg.StrOpt(
        'host',
        default='127.0.0.1',
        help='Etcd service host address.'
    ),
    cfg.PortOpt(
        'port',
        default=2379,
        help='Etcd service port.'
    ),
    cfg.StrOpt(
        'protocol',
        default='https',
        choices=['http', 'https'],
        help='Etcd connection protocol.'
    ),
    cfg.StrOpt(
        'ca_cert',
        default='/etc/qinling/pki/etcd/ca.crt',
        help='Path to CA certificate file to use to securely '
             'connect to etcd server.'
    ),
    cfg.StrOpt(
        'cert_file',
        default='/etc/qinling/pki/etcd/qinling-etcd-client.crt',
        help='Path to client certificate file to use to securely '
             'connect to etcd server.'
    ),
    cfg.StrOpt(
        'cert_key',
        default='/etc/qinling/pki/etcd/qinling-etcd-client.key',
        help='Path to client certificate key file to use to securely '
             'connect to etcd server.'
    ),
]

RLIMITS_GROUP = 'resource_limits'
rlimits_opts = [
    cfg.IntOpt(
        'default_cpu',
        default=100,
        help='Default cpu resource(unit: millicpu).'
    ),
    cfg.IntOpt(
        'min_cpu',
        default=100,
        help='Minimum cpu resource(unit: millicpu).'
    ),
    cfg.IntOpt(
        'max_cpu',
        default=300,
        help='Maximum cpu resource(unit: millicpu).'
    ),
    cfg.IntOpt(
        'default_memory',
        default=33554432,
        help='Default memory resource(unit: bytes).'
    ),
    cfg.IntOpt(
        'min_memory',
        default=33554432,
        help='Minimum memory resource(unit: bytes).'
    ),
    cfg.IntOpt(
        'max_memory',
        default=134217728,
        help='Maximum memory resource(unit: bytes).'
    ),
    cfg.IntOpt(
        'default_timeout',
        default=5,
        help='Default function execution timeout(unit: seconds)'
    ),
    cfg.IntOpt(
        'min_timeout',
        default=1,
        help='Minimum function execution timeout(unit: seconds).'
    ),
    cfg.IntOpt(
        'max_timeout',
        default=300,
        help='Maximum function execution timeout(unit: seconds).'
    ),
]


def list_opts():
    keystone_middleware_opts = auth_token.list_opts()
    keystone_loading_opts = [(
        'keystone_authtoken', loading.get_auth_plugin_conf_options('password')
    )]

    qinling_opts = [
        (API_GROUP, api_opts),
        (PECAN_GROUP, pecan_opts),
        (ENGINE_GROUP, engine_opts),
        (STORAGE_GROUP, storage_opts),
        (KUBERNETES_GROUP, kubernetes_opts),
        (ETCD_GROUP, etcd_opts),
        (RLIMITS_GROUP, rlimits_opts),
        (None, [launch_opt]),
        (None, default_opts),
    ]

    return keystone_middleware_opts + keystone_loading_opts + qinling_opts


def parse_args(args=None, usage=None, default_config_files=None):
    CLI_OPTS = [launch_opt]
    CONF.register_cli_opts(CLI_OPTS)

    for group, options in list_opts():
        CONF.register_opts(list(options), group)

    _DEFAULT_LOG_LEVELS = [
        'eventlet.wsgi.server=WARN',
        'oslo_service.periodic_task=INFO',
        'oslo_service.loopingcall=INFO',
        'oslo_db=WARN',
        'oslo_concurrency.lockutils=WARN',
        'kubernetes.client.rest=%s' % CONF.kubernetes.log_devel,
        'keystoneclient=INFO',
        'requests.packages.urllib3.connectionpool=CRITICAL',
        'urllib3.connectionpool=CRITICAL',
        'cotyledon=INFO',
        'futurist.periodics=WARN'
    ]
    default_log_levels = log.get_default_log_levels()
    default_log_levels.extend(_DEFAULT_LOG_LEVELS)
    log.set_defaults(default_log_levels=default_log_levels)
    log.register_options(CONF)

    CONF(
        args=args,
        project='qinling',
        version=version,
        usage=usage,
        default_config_files=default_config_files
    )


def set_config_defaults():
    """This method updates all configuration default values."""
    set_cors_middleware_defaults()


def set_cors_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id',
                       'X-Project-Id',
                       'X-User-Name',
                       'X-Project-Name'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH'],
        expose_headers=['X-Auth-Token',
                        'X-Subject-Token',
                        'X-Service-Token',
                        'X-Project-Id',
                        'X-User-Name',
                        'X-Project-Name']
    )



# Copyright 2018 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json
import mock

from qinling import context
from qinling.db import api as db_api
from qinling.tests.unit.api import base
from qinling.utils import constants


class TestWebhookController(base.APITest):
    def setUp(self):
        super(TestWebhookController, self).setUp()
        db_func = self.create_function()
        self.func_id = db_func.id

    def test_crud(self):
        # Create
        body = {
            'function_id': self.func_id,
            'description': 'webhook test'
        }
        resp = self.app.post_json('/v1/webhooks', body)
        self.assertEqual(201, resp.status_int)
        webhook_id = resp.json.get('id')
        self.assertIn(self.qinling_endpoint, resp.json.get('webhook_url'))

        # Get
        resp = self.app.get('/v1/webhooks/%s' % webhook_id)
        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, body)

        # List
        resp = self.app.get('/v1/webhooks')
        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(
            resp.json['webhooks'], id=webhook_id
        )
        self._assertDictContainsSubset(actual, body)

        # Update
        resp = self.app.put_json(
            '/v1/webhooks/%s' % webhook_id,
            {'description': 'webhook test update'}
        )
        self.assertEqual(200, resp.status_int)

        expected = {
            'function_id': self.func_id,
            'description': 'webhook test update'
        }
        resp = self.app.get('/v1/webhooks/%s' % webhook_id)
        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, expected)

        # Delete
        resp = self.app.delete('/v1/webhooks/%s' % webhook_id)
        self.assertEqual(204, resp.status_int)
        resp = self.app.get('/v1/webhooks/%s' % webhook_id, expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_create_with_version(self):
        db_api.increase_function_version(self.func_id, 0)

        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'description': 'webhook test'
        }
        resp = self.app.post_json('/v1/webhooks', body)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(1, resp.json.get("function_version"))
        self.assertIsNone(resp.json.get("function_alias"))

    def test_create_with_alias(self):
        db_api.increase_function_version(self.func_id, 0)
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'name': name
        }
        db_api.create_function_alias(**body)

        webhook_body = {
            'function_alias': name,
            'description': 'webhook test'
        }
        resp = self.app.post_json('/v1/webhooks', webhook_body)

        self.assertEqual(201, resp.status_int)
        self.assertEqual(name, resp.json.get('function_alias'))
        self.assertIsNone(resp.json.get("function_id"))
        self.assertIsNone(resp.json.get("function_version"))

    def test_create_with_invalid_alias(self):
        body = {
            'function_alias': 'fake_alias',
            'description': 'webhook test'
        }

        resp = self.app.post_json('/v1/webhooks', body, expect_errors=True)

        self.assertEqual(404, resp.status_int)

    def test_create_without_required_params(self):
        resp = self.app.post(
            '/v1/webhooks',
            params={},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)

    def test_update_with_version(self):
        db_api.increase_function_version(self.func_id, 0)
        webhook = self.create_webhook(self.func_id)

        self.assertIsNone(webhook.function_version)

        resp = self.app.put_json(
            '/v1/webhooks/%s' % webhook.id,
            {'function_version': 1}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, resp.json.get("function_version"))
        self.assertIsNone(resp.json.get("function_alias"))

    def test_update_only_description(self):
        db_api.increase_function_version(self.func_id, 0)
        webhook = self.create_webhook(self.func_id, function_version=1)

        self.assertEqual(1, webhook.function_version)

        resp = self.app.put_json(
            '/v1/webhooks/%s' % webhook.id,
            {'description': 'updated description'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(1, resp.json.get("function_version"))
        self.assertEqual('updated description', resp.json.get("description"))

    def test_update_function_alias_1(self):
        # Create webhook using function alias
        db_api.increase_function_version(self.func_id, 0)
        name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'name': name
        }
        db_api.create_function_alias(**body)
        webhook = self.create_webhook(function_alias=name)

        db_api.increase_function_version(self.func_id, 1)
        new_name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 2,
            'name': new_name
        }
        db_api.create_function_alias(**body)

        # Update webhook with the new alias
        resp = self.app.put_json(
            '/v1/webhooks/%s' % webhook.id,
            {'function_alias': new_name}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(new_name, resp.json.get("function_alias"))
        self.assertIsNone(resp.json.get("function_id"))
        self.assertIsNone(resp.json.get("function_version"))

    def test_update_function_alias_2(self):
        # Create webhook using function id
        db_api.increase_function_version(self.func_id, 0)
        webhook = self.create_webhook(function_id=self.func_id,
                                      function_version=1)

        db_api.increase_function_version(self.func_id, 1)
        alias_name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 2,
            'name': alias_name
        }
        db_api.create_function_alias(**body)

        # Update webhook with function alias
        resp = self.app.put_json(
            '/v1/webhooks/%s' % webhook.id,
            {'function_alias': alias_name}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(alias_name, resp.json.get("function_alias"))
        self.assertIsNone(resp.json.get("function_id"))
        self.assertIsNone(resp.json.get("function_version"))

    @mock.patch("qinling.utils.openstack.keystone.create_trust_context")
    @mock.patch("qinling.utils.executions.create_execution")
    def test_invoke_with_function_id(self, mock_create_execution,
                                     mock_create_context):
        exec_mock = mock_create_execution.return_value
        exec_mock.id = "fake_id"
        webhook = self.create_webhook(function_id=self.func_id)

        resp = self.app.post_json('/v1/webhooks/%s/invoke' % webhook.id, {})
        context.set_ctx(self.ctx)

        self.assertEqual(202, resp.status_int)

        params = {
            'function_id': self.func_id,
            'function_version': None,
            'sync': False,
            'input': json.dumps({}),
            'description': constants.EXECUTION_BY_WEBHOOK % webhook.id
        }
        mock_create_execution.assert_called_once_with(mock.ANY, params)

    @mock.patch("qinling.utils.openstack.keystone.create_trust_context")
    @mock.patch("qinling.utils.executions.create_execution")
    def test_invoke_with_function_alias(self, mock_create_execution,
                                        mock_create_context):
        exec_mock = mock_create_execution.return_value
        exec_mock.id = "fake_id"

        db_api.increase_function_version(self.func_id, 0)
        alias_name = self.rand_name(name="alias", prefix=self.prefix)
        body = {
            'function_id': self.func_id,
            'function_version': 1,
            'name': alias_name
        }
        db_api.create_function_alias(**body)
        webhook = self.create_webhook(function_alias=alias_name)

        resp = self.app.post_json('/v1/webhooks/%s/invoke' % webhook.id, {})
        context.set_ctx(self.ctx)

        self.assertEqual(202, resp.status_int)

        params = {
            'function_id': self.func_id,
            'function_version': 1,
            'sync': False,
            'input': json.dumps({}),
            'description': constants.EXECUTION_BY_WEBHOOK % webhook.id
        }
        mock_create_execution.assert_called_once_with(mock.ANY, params)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import datetime

import croniter
from dateutil import parser
from oslo_utils import timeutils
import six

from qinling import exceptions as exc


def validate_next_time(next_execution_time):
    next_time = next_execution_time
    if isinstance(next_execution_time, six.string_types):
        try:
            # We need naive datetime object.
            next_time = parser.parse(next_execution_time, ignoretz=True)
        except ValueError as e:
            raise exc.InputException(str(e))

    valid_min_time = timeutils.utcnow() + datetime.timedelta(0, 60)
    if valid_min_time > next_time:
        raise exc.InputException(
            'Execution time must be at least 1 minute in the future.'
        )

    return next_time


def validate_pattern(pattern):
    try:
        croniter.croniter(pattern)
    except (ValueError, KeyError):
        raise exc.InputException(
            'The specified pattern is not valid: {}'.format(pattern)
        )


def validate_job(params):
    first_time = params.get('first_execution_time')
    pattern = params.get('pattern')
    count = params.get('count')
    start_time = timeutils.utcnow()

    if not (first_time or pattern):
        raise exc.InputException(
            'pattern or first_execution_time must be specified.'
        )

    if first_time:
        first_time = validate_next_time(first_time)
        if not pattern and count and count > 1:
            raise exc.InputException(
                'pattern must be provided if count is greater than 1.'
            )

        next_time = first_time
        if not (pattern or count):
            count = 1
    if pattern:
        validate_pattern(pattern)

        if first_time:
            start_time = first_time - datetime.timedelta(minutes=1)

        next_time = croniter.croniter(pattern, start_time).get_next(
            datetime.datetime
        )
        first_time = next_time

    return first_time, next_time, count


def get_next_execution_time(pattern, start_time):
    return croniter.croniter(pattern, start_time).get_next(
        datetime.datetime
    )



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime

import croniter
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
from pecan import rest
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from qinling.api import access_control as acl
from qinling.api.controllers.v1 import resources
from qinling.api.controllers.v1 import types
from qinling import context
from qinling.db import api as db_api
from qinling import exceptions as exc
from qinling import status
from qinling.utils import jobs
from qinling.utils import rest_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

UPDATE_ALLOWED = set(['name', 'function_input', 'status', 'pattern',
                      'next_execution_time'])


class JobsController(rest.RestController):
    type = 'job'

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Job,
        body=resources.Job,
        status_code=201
    )
    def post(self, job):
        """Creates a new job."""
        params = job.to_dict()
        if not (params.get("function_id") or params.get("function_alias")):
            raise exc.InputException(
                'Either function_alias or function_id must be provided.'
            )

        LOG.info("Creating %s, params: %s", self.type, params)

        # Check the input params.
        first_time, next_time, count = jobs.validate_job(params)

        version = params.get('function_version', 0)
        function_alias = params.get('function_alias')

        if function_alias:
            # Check if the alias exists.
            db_api.get_function_alias(function_alias)
        else:
            # Check the function(version) exists.
            db_api.get_function(params['function_id'])
            if version > 0:
                # Check if the version exists.
                db_api.get_function_version(params['function_id'], version)

        values = {
            'name': params.get('name'),
            'pattern': params.get('pattern'),
            'first_execution_time': first_time,
            'next_execution_time': next_time,
            'count': count,
            'function_alias': function_alias,
            'function_id': params.get("function_id"),
            'function_version': version,
            'function_input': params.get('function_input'),
            'status': status.RUNNING
        }
        db_job = db_api.create_job(values)

        return resources.Job.from_db_obj(db_job)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        LOG.info("Delete resource.", resource={'type': self.type, 'id': id})
        return db_api.delete_job(id)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Job, types.uuid)
    def get(self, id):
        LOG.info("Get resource.", resource={'type': self.type, 'id': id})
        job_db = db_api.get_job(id)

        return resources.Job.from_db_obj(job_db)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.Jobs, bool, wtypes.text)
    def get_all(self, all_projects=False, project_id=None):
        project_id, all_projects = rest_utils.get_project_params(
            project_id, all_projects
        )
        if all_projects:
            acl.enforce('job:get_all:all_projects', context.get_ctx())

        filters = rest_utils.get_filters(
            project_id=project_id,
        )
        LOG.info("Get all %ss. filters=%s", self.type, filters)
        db_jobs = db_api.get_jobs(insecure=all_projects, **filters)
        jobs = [resources.Job.from_db_obj(db_model)
                for db_model in db_jobs]

        return resources.Jobs(jobs=jobs)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(
        resources.Job,
        types.uuid,
        body=resources.Job
    )
    def put(self, id, job):
        """Update job definition.

        1. Can not update a finished job.
        2. Can not change job type.
        3. Allow to pause a one-shot job and resume before its first execution
           time.
        """
        values = {}
        for key in UPDATE_ALLOWED:
            if job.to_dict().get(key) is not None:
                values.update({key: job.to_dict()[key]})

        LOG.info('Update resource, params: %s', values,
                 resource={'type': self.type, 'id': id})

        new_status = values.get('status')
        pattern = values.get('pattern')
        next_execution_time = values.get('next_execution_time')

        job_db = db_api.get_job(id)

        if job_db.status in [status.DONE, status.CANCELLED]:
            raise exc.InputException('Can not update a finished job.')

        if pattern:
            if not job_db.pattern:
                raise exc.InputException('Can not change job type.')
            jobs.validate_pattern(pattern)
        elif pattern == '' and job_db.pattern:
            raise exc.InputException('Can not change job type.')

        valid_states = [status.RUNNING, status.CANCELLED, status.PAUSED]
        if new_status and new_status not in valid_states:
            raise exc.InputException('Invalid status.')

        if next_execution_time:
            values['next_execution_time'] = jobs.validate_next_time(
                next_execution_time
            )
        elif (job_db.status == status.PAUSED and
              new_status == status.RUNNING):
            p = job_db.pattern or pattern

            if not p:
                # Check if the next execution time for one-shot job is still
                # valid.
                jobs.validate_next_time(job_db.next_execution_time)
            else:
                # Update next_execution_time for recurring job.
                values['next_execution_time'] = croniter.croniter(
                    p, timeutils.utcnow()
                ).get_next(datetime.datetime)

        updated_job = db_api.update_job(id, values)
        return resources.Job.from_db_obj(updated_job)



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import mock

from qinling.db import api as db_api
from qinling import status
from qinling.tests.unit.api import base
from qinling.tests.unit import base as test_base


class TestRuntimeController(base.APITest):
    def setUp(self):
        super(TestRuntimeController, self).setUp()

        # Insert a runtime record in db. The data will be removed in db clean
        # up.
        self.db_runtime = self.create_runtime()
        self.runtime_id = self.db_runtime.id

    def test_get(self):
        resp = self.app.get('/v1/runtimes/%s' % self.runtime_id)

        expected = {
            'id': self.runtime_id,
            "image": self.db_runtime.image,
            "name": self.db_runtime.name,
            "project_id": test_base.DEFAULT_PROJECT_ID,
            "status": status.AVAILABLE,
            "is_public": True,
        }

        self.assertEqual(200, resp.status_int)
        self._assertDictContainsSubset(resp.json, expected)

    def test_get_all(self):
        resp = self.app.get('/v1/runtimes')

        expected = {
            'id': self.runtime_id,
            "image": self.db_runtime.image,
            "name": self.db_runtime.name,
            "project_id": test_base.DEFAULT_PROJECT_ID,
            "status": status.AVAILABLE,
            "is_public": True,
        }

        self.assertEqual(200, resp.status_int)
        actual = self._assert_single_item(
            resp.json['runtimes'], id=self.runtime_id
        )
        self._assertDictContainsSubset(actual, expected)

    @mock.patch('qinling.rpc.EngineClient.create_runtime')
    def test_post(self, mock_create_time):
        body = {
            'name': self.rand_name('runtime', prefix=self.prefix),
            'image': self.rand_name('image', prefix=self.prefix),
        }
        resp = self.app.post_json('/v1/runtimes', body)

        self.assertEqual(201, resp.status_int)

        body.update({"trusted": True})
        self._assertDictContainsSubset(resp.json, body)

        mock_create_time.assert_called_once_with(resp.json['id'])

    @mock.patch('qinling.rpc.EngineClient.create_runtime')
    def test_post_without_image(self, mock_create_time):
        body = {
            'name': self.rand_name('runtime', prefix=self.prefix),
        }
        resp = self.app.post_json('/v1/runtimes', body, expect_errors=True)

        self.assertEqual(400, resp.status_int)
        mock_create_time.assert_not_called()

    @mock.patch('qinling.rpc.EngineClient.delete_runtime')
    def test_delete(self, mock_delete_runtime):
        resp = self.app.delete('/v1/runtimes/%s' % self.runtime_id)

        self.assertEqual(204, resp.status_int)
        mock_delete_runtime.assert_called_once_with(self.runtime_id)

    @mock.patch('qinling.rpc.EngineClient.delete_runtime')
    def test_delete_runtime_with_function_associated(self,
                                                     mock_delete_runtime):
        self.create_function(self.runtime_id)
        resp = self.app.delete(
            '/v1/runtimes/%s' % self.runtime_id, expect_errors=True
        )

        self.assertEqual(403, resp.status_int)
        mock_delete_runtime.assert_not_called()

    def test_put_name(self):
        resp = self.app.put_json(
            '/v1/runtimes/%s' % self.runtime_id, {'name': 'new_name'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual('new_name', resp.json['name'])

    def test_put_image_runtime_not_available(self):
        db_runtime = db_api.create_runtime(
            {
                'name': self.rand_name('runtime', prefix=self.prefix),
                'image': self.rand_name('image', prefix=self.prefix),
                'project_id': test_base.DEFAULT_PROJECT_ID,
                'status': status.CREATING
            }
        )
        runtime_id = db_runtime.id

        resp = self.app.put_json(
            '/v1/runtimes/%s' % runtime_id, {'image': 'new_image'},
            expect_errors=True
        )

        self.assertEqual(409, resp.status_int)

    @mock.patch('qinling.rpc.EngineClient.update_runtime')
    def test_put_image(self, mock_update_runtime):
        resp = self.app.put_json(
            '/v1/runtimes/%s' % self.runtime_id, {'image': 'new_image'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual('new_image', resp.json['image'])
        mock_update_runtime.assert_called_once_with(
            self.runtime_id,
            image='new_image',
            pre_image=self.db_runtime.image
        )

    @mock.patch('qinling.utils.etcd_util.get_service_url')
    @mock.patch('qinling.rpc.EngineClient.update_runtime')
    def test_put_image_not_allowed(self, mock_update_runtime, mock_etcd_url):
        mock_etcd_url.return_value = True
        function_id = self.create_function(self.runtime_id).id

        resp = self.app.put_json(
            '/v1/runtimes/%s' % self.runtime_id, {'image': 'new_image'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)
        mock_update_runtime.assert_not_called()
        mock_etcd_url.assert_called_once_with(function_id)

    @mock.patch('qinling.rpc.EngineClient.get_runtime_pool')
    def test_get_runtime_pool(self, mock_get_pool):
        mock_get_pool.return_value = {"total": 3, "available": 2}

        resp = self.app.get('/v1/runtimes/%s/pool' % self.runtime_id)

        expected = {
            "capacity": {
                "available": 2,
                "total": 3
            },
            "name": self.runtime_id
        }

        self.assertEqual(200, resp.status_int)
        self.assertEqual(expected, resp.json)



# Copyright 2019 - Ormuco Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Make function id nullable for webhooks table

Revision ID: 011
Revises: 010
"""

revision = '011'
down_revision = '010'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column(
        'webhooks',
        'function_id',
        existing_type=sa.String(length=36),
        nullable=True
    )
    op.alter_column(
        'webhooks',
        'function_version',
        existing_type=sa.Integer,
        nullable=True
    )



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import os
import pkg_resources
import tempfile
import zipfile

from tempest import config
from tempest.lib.common.utils import data_utils
from tempest import test
import tenacity

CONF = config.CONF


class BaseQinlingTest(test.BaseTestCase):
    credentials = ('admin', 'primary', 'alt')
    create_runtime = True
    image = CONF.qinling.python_runtime_image

    @classmethod
    def skip_checks(cls):
        super(BaseQinlingTest, cls).skip_checks()

        if not CONF.service_available.qinling:
            raise cls.skipException("Qinling service is not available.")

    @classmethod
    def setup_clients(cls):
        super(BaseQinlingTest, cls).setup_clients()

        cls.client = cls.os_primary.qinling.QinlingClient()
        cls.alt_client = cls.os_alt.qinling.QinlingClient()
        cls.admin_client = cls.os_admin.qinling.QinlingClient()

    @classmethod
    def resource_setup(cls):
        super(BaseQinlingTest, cls).resource_setup()

        if cls.create_runtime:
            cls.runtime_id = None
            name = data_utils.rand_name('runtime', prefix=cls.name_prefix)
            _, body = cls.admin_client.create_runtime(cls.image, name)
            cls.runtime_id = body['id']

    @classmethod
    def resource_cleanup(cls):
        if cls.create_runtime and cls.runtime_id:
            cls.admin_client.delete_resource(
                'runtimes', cls.runtime_id,
                ignore_notfound=True
            )

        super(BaseQinlingTest, cls).resource_cleanup()

    @tenacity.retry(
        wait=tenacity.wait_fixed(3),
        stop=tenacity.stop_after_attempt(20),
        retry=tenacity.retry_if_exception_type(AssertionError)
    )
    def wait_runtime_available(self, id):
        resp, body = self.client.get_resource('runtimes', id)

        self.assertEqual(200, resp.status)
        self.assertEqual('available', body['status'])

    @tenacity.retry(
        wait=tenacity.wait_fixed(3),
        stop=tenacity.stop_after_attempt(10),
        retry=tenacity.retry_if_exception_type(AssertionError)
    )
    def wait_execution_success(self, id):
        resp, body = self.client.get_resource('executions', id)

        self.assertEqual(200, resp.status)
        self.assertEqual('success', body['status'])

    @tenacity.retry(
        wait=tenacity.wait_fixed(10),
        stop=tenacity.stop_after_attempt(12),
        retry=tenacity.retry_if_exception_type(AssertionError)
    )
    def wait_job_done(self, id):
        resp, body = self.client.get_resource('jobs', id)

        self.assertEqual(200, resp.status)
        self.assertEqual('done', body['status'])

    def create_package(self, name="python/test_python_basic.py"):
        file_path = pkg_resources.resource_filename(
            'qinling_tempest_plugin',
            "functions/%s" % name
        )
        base_name, extension = os.path.splitext(file_path)
        module_name = os.path.basename(base_name)
        temp_dir = tempfile.mkdtemp()
        zip_file = os.path.join(temp_dir, '%s.zip' % module_name)

        if not os.path.isfile(zip_file):
            zf = zipfile.ZipFile(zip_file, mode='w')
            try:
                zf.write(file_path, '%s%s' % (module_name, extension))
            finally:
                zf.close()

        self.addCleanup(os.rmdir, temp_dir)
        self.addCleanup(os.remove, zip_file)
        return zip_file

    def create_function(self, package_path=None, image=None,
                        md5sum=None, timeout=None):
        function_name = data_utils.rand_name(
            'function',
            prefix=self.name_prefix
        )

        if not image:
            if not package_path:
                package_path = self.create_package()

            code = {"source": "package"}
            if md5sum:
                code.update({"md5sum": md5sum})
            base_name, _ = os.path.splitext(package_path)
            module_name = os.path.basename(base_name)
            with open(package_path, 'rb') as package_data:
                resp, body = self.client.create_function(
                    code,
                    self.runtime_id,
                    name=function_name,
                    package_data=package_data,
                    entry='%s.main' % module_name,
                    timeout=timeout
                )
        else:
            resp, body = self.client.create_function(
                {"source": "image", "image": image},
                None,
                name=function_name,
            )

        self.assertEqual(201, resp.status_code)
        function_id = body['id']
        self.addCleanup(self.client.delete_resource, 'functions',
                        function_id, ignore_notfound=True)

        return function_id

    def update_function_package(self, function_id, function_path):
        package_path = self.create_package(name=function_path)
        base_name, _ = os.path.splitext(package_path)
        module_name = os.path.basename(base_name)

        with open(package_path, 'rb') as package_data:
            resp, _ = self.client.update_function(
                function_id,
                package_data=package_data,
                entry='%s.main' % module_name
            )

        self.assertEqual(200, resp.status_code)

    def create_webhook(self, function_id=None, function_alias=None,
                       version=0):
        if function_alias:
            resp, body = self.client.create_webhook(
                function_alias=function_alias
            )
        else:
            if not function_id:
                function_id = self.create_function()
            resp, body = self.client.create_webhook(
                function_id,
                version=version
            )
        self.assertEqual(201, resp.status)

        webhook_id = body['id']
        self.addCleanup(self.client.delete_resource, 'webhooks',
                        webhook_id, ignore_notfound=True)

        return webhook_id, body['webhook_url']

    def create_job(self, function_id=None, function_alias=None, version=0,
                   first_execution_time=None):
        if function_alias:
            resp, body = self.client.create_job(
                function_alias=function_alias,
                first_execution_time=first_execution_time
            )
        else:
            if not function_id:
                function_id = self.create_function()
            resp, body = self.client.create_job(
                function_id,
                version=version,
                first_execution_time=first_execution_time
            )
        self.assertEqual(201, resp.status)
        job_id = body['id']

        self.addCleanup(self.client.delete_resource, 'jobs',
                        job_id, ignore_notfound=True)

        return job_id

    def create_function_version(self, function_id=None):
        if not function_id:
            function_id = self.create_function()

        resp, body = self.client.create_function_version(function_id)
        self.assertEqual(201, resp.status)

        version = body['version_number']

        self.addCleanup(self.client.delete_function_version, function_id,
                        version, ignore_notfound=True)

        return version

    def create_execution(self, function_id=None, alias_name=None, version=0,
                         input=None):
        if alias_name:
            resp, body = self.client.create_execution(alias_name=alias_name,
                                                      input=input)
        else:
            resp, body = self.client.create_execution(function_id,
                                                      version=version,
                                                      input=input)

        self.assertEqual(201, resp.status)

        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        self.assertEqual('success', body['status'])

        return execution_id

    def create_function_alias(self, function_id=None, function_version=0):
        name = data_utils.rand_name(name="alias", prefix=self.name_prefix)
        if not function_id:
            function_id = self.create_function()

        resp, body = self.client.create_function_alias(name,
                                                       function_id,
                                                       function_version)

        self.assertEqual(201, resp.status)

        alias_name = body['name']
        self.addCleanup(self.client.delete_function_alias, alias_name,
                        ignore_notfound=True)

        return alias_name



# Copyright 2017 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import json
import os
import time

import jinja2
from oslo_log import log as logging
import requests
import tenacity
import yaml

from qinling.engine import utils
from qinling import exceptions as exc
from qinling.orchestrator import base
from qinling.orchestrator.kubernetes import utils as k8s_util
from qinling.utils import common


LOG = logging.getLogger(__name__)

TEMPLATES_DIR = (os.path.dirname(os.path.realpath(__file__)) + '/templates/')


class KubernetesManager(base.OrchestratorBase):
    def __init__(self, conf, qinling_endpoint):
        self.conf = conf
        self.qinling_endpoint = qinling_endpoint

        clients = k8s_util.get_k8s_clients(self.conf)
        self.v1 = clients['v1']
        self.v1extension = clients['v1extension']
        # self.apps_v1 = clients['apps_v1']

        # Create namespace if not exists
        self._ensure_namespace()

        # Get templates.
        template_loader = jinja2.FileSystemLoader(
            searchpath=os.path.dirname(TEMPLATES_DIR)
        )
        jinja_env = jinja2.Environment(
            loader=template_loader, autoescape=True, trim_blocks=True,
            lstrip_blocks=True
        )
        self.deployment_template = jinja_env.get_template('deployment.j2')
        self.service_template = jinja_env.get_template('service.j2')
        self.pod_template = jinja_env.get_template('pod.j2')

        # Refer to
        # http://docs.python-requests.org/en/master/user/advanced/#session-objects
        self.session = requests.Session()

    def _ensure_namespace(self):
        ret = self.v1.list_namespace()
        cur_names = [i.metadata.name for i in ret.items]

        if self.conf.kubernetes.namespace not in cur_names:
            LOG.info('Creating namespace: %s', self.conf.kubernetes.namespace)

            namespace_body = {
                'apiVersion': 'v1',
                'kind': 'Namespace',
                'metadata': {
                    'name': self.conf.kubernetes.namespace,
                    'labels': {
                        'name': self.conf.kubernetes.namespace
                    }
                },
            }

            self.v1.create_namespace(namespace_body)

            LOG.info('Namespace %s created.', self.conf.kubernetes.namespace)

    @tenacity.retry(
        wait=tenacity.wait_fixed(2),
        stop=tenacity.stop_after_delay(600),
        reraise=True,
        retry=tenacity.retry_if_exception_type(exc.OrchestratorException)
    )
    def _wait_deployment_available(self, name):
        ret = self.v1extension.read_namespaced_deployment(
            name,
            self.conf.kubernetes.namespace
        )

        if (not ret.status.replicas or
                ret.status.replicas != ret.status.available_replicas):
            raise exc.OrchestratorException('Deployment %s not ready.' % name)

    def get_pool(self, name):
        total = 0
        available = 0

        try:
            ret = self.v1extension.read_namespaced_deployment(
                name,
                namespace=self.conf.kubernetes.namespace
            )
        except Exception:
            raise exc.RuntimeNotFoundException()

        if not ret.status.replicas:
            return {"total": total, "available": available}

        total = ret.status.replicas

        labels = {'runtime_id': name}
        selector = common.convert_dict_to_string(labels)
        ret = self.v1.list_namespaced_pod(
            self.conf.kubernetes.namespace,
            label_selector='!function_id,%s' % selector
        )
        available = len(ret.items)

        return {"total": total, "available": available}

    def create_pool(self, name, image, trusted=True):
        deployment_body = self.deployment_template.render(
            {
                "name": name,
                "labels": {'runtime_id': name},
                "replicas": self.conf.kubernetes.replicas,
                "container_name": 'worker',
                "image": image,
                "sidecar_image": self.conf.engine.sidecar_image,
                "trusted": str(trusted).lower()
            }
        )

        LOG.info(
            "Creating deployment for runtime %s: \n%s", name, deployment_body
        )

        self.v1extension.create_namespaced_deployment(
            body=yaml.safe_load(deployment_body),
            namespace=self.conf.kubernetes.namespace,
            async_req=False
        )

        self._wait_deployment_available(name)

        LOG.info("Deployment for runtime %s created.", name)

    def delete_pool(self, name):
        """Delete all resources belong to the deployment."""
        LOG.info("Deleting deployment %s", name)

        labels = {'runtime_id': name}
        selector = common.convert_dict_to_string(labels)

        self.v1extension.delete_collection_namespaced_replica_set(
            self.conf.kubernetes.namespace,
            label_selector=selector
        )
        LOG.info("ReplicaSets in deployment %s deleted.", name)

        ret = self.v1.list_namespaced_service(
            self.conf.kubernetes.namespace, label_selector=selector
        )
        names = [i.metadata.name for i in ret.items]
        for svc_name in names:
            self.v1.delete_namespaced_service(
                svc_name,
                self.conf.kubernetes.namespace
            )
        LOG.info("Services in deployment %s deleted.", name)

        self.v1extension.delete_collection_namespaced_deployment(
            self.conf.kubernetes.namespace,
            label_selector=selector,
            field_selector='metadata.name=%s' % name
        )
        # Should delete pods after deleting deployment to avoid pods are
        # recreated by k8s.
        self.v1.delete_collection_namespaced_pod(
            self.conf.kubernetes.namespace,
            label_selector=selector
        )
        LOG.info("Pods in deployment %s deleted.", name)
        LOG.info("Deployment %s deleted.", name)

    @tenacity.retry(
        wait=tenacity.wait_fixed(5),
        stop=tenacity.stop_after_delay(600),
        reraise=True,
        retry=tenacity.retry_if_exception_type(exc.OrchestratorException)
    )
    def _wait_for_upgrade(self, deploy_name):
        ret = self.v1extension.read_namespaced_deployment(
            deploy_name,
            self.conf.kubernetes.namespace
        )
        if ret.status.unavailable_replicas is not None:
            raise exc.OrchestratorException("Deployment %s upgrade not "
                                            "ready." % deploy_name)

    def update_pool(self, name, image=None):
        """Deployment rolling-update.

        Return True if successful, otherwise return False after rolling back.
        """
        LOG.info('Start to do rolling-update deployment %s', name)

        body = {
            'spec': {
                'template': {
                    'spec': {
                        'containers': [
                            {
                                'name': 'worker',
                                'image': image
                            }
                        ]
                    }
                }
            }
        }
        self.v1extension.patch_namespaced_deployment(
            name, self.conf.kubernetes.namespace, body
        )

        try:
            time.sleep(10)
            self._wait_for_upgrade(name)
        except exc.OrchestratorException:
            LOG.warn("Timeout when waiting for the deployment %s upgrade, "
                     "Start to roll back.", name)

            body = {"rollbackTo": {"revision": 0}}
            try:
                self.v1extension.create_namespaced_deployment_rollback(
                    name, self.conf.kubernetes.namespace, body
                )
            except Exception:
                # TODO(lxkong): remove the exception catch until kubernetes
                # python lib has a new release. Refer to
                # https://github.com/kubernetes-client/python/issues/491
                pass

            return False

        return True

    def _choose_available_pods(self, labels, count=1, function_id=None,
                               function_version=0):
        # If there is already a pod for function, reuse it.
        if function_id:
            ret = self.v1.list_namespaced_pod(
                self.conf.kubernetes.namespace,
                label_selector='function_id=%s,function_version=%s' %
                               (function_id, function_version)
            )
            if len(ret.items) >= count:
                LOG.debug(
                    "Function %s(version %s) already associates to a pod with "
                    "at least %d worker(s). ",
                    function_id, function_version, count
                )
                return ret.items[:count]

        selector = common.convert_dict_to_string(labels)
        ret = self.v1.list_namespaced_pod(
            self.conf.kubernetes.namespace,
            label_selector='!function_id,%s' % selector
        )

        if len(ret.items) < count:
            return []

        return ret.items[-count:]

    def _prepare_pod(self, pod, deployment_name, function_id, version,
                     labels=None):
        """Pod preparation.

        1. Update pod labels.
        2. Expose service.
        """
        pod_name = pod.metadata.name
        labels = labels or {}

        LOG.info(
            'Prepare pod %s in deployment %s for function %s(version %s)',
            pod_name, deployment_name, function_id, version
        )

        # Update pod label.
        pod_labels = self._update_pod_label(
            pod,
            # pod label value should be string
            {'function_id': function_id, 'function_version': str(version)}
        )

        # Create service for the chosen pod.
        service_name = "service-%s-%s" % (function_id, version)
        labels.update(
            {'function_id': function_id, 'function_version': str(version)}
        )

        # TODO(kong): Make the service type configurable.
        service_body = self.service_template.render(
            {
                "service_name": service_name,
                "labels": labels,
                "selector": pod_labels
            }
        )
        try:
            ret = self.v1.create_namespaced_service(
                self.conf.kubernetes.namespace, yaml.safe_load(service_body)
            )
            LOG.debug(
                'Service created for pod %s, service name: %s',
                pod_name, service_name
            )
        except Exception as e:
            # Service already exists
            if e.status == 409:
                LOG.debug(
                    'Service already exists for pod %s, service name: %s',
                    pod_name, service_name
                )
                time.sleep(1)
                ret = self.v1.read_namespaced_service(
                    service_name, self.conf.kubernetes.namespace
                )
            else:
                raise

        # Get external ip address for an arbitrary node.
        node_port = ret.spec.ports[0].node_port
        nodes = self.v1.list_node()
        addresses = nodes.items[0].status.addresses
        node_ip = None
        for addr in addresses:
            if addr.type == 'ExternalIP':
                node_ip = addr.address

        # FIXME: test purpose using minikube
        if not node_ip:
            for addr in addresses:
                if addr.type == 'InternalIP':
                    node_ip = addr.address

        pod_service_url = 'http://%s:%s' % (node_ip, node_port)

        return pod_name, pod_service_url

    def _create_pod(self, image, rlimit, pod_name, labels, input):
        """Create pod for image type function."""
        if not input:
            input_list = []
        elif isinstance(input, dict) and input.get('__function_input'):
            input_list = input.get('__function_input').split()
        else:
            input_list = list(json.loads(input))

        pod_body = self.pod_template.render(
            {
                "pod_name": pod_name,
                "labels": labels,
                "pod_image": image,
                "input": input_list,
                "req_cpu": str(rlimit['cpu']),
                "limit_cpu": str(rlimit['cpu']),
                "req_memory": str(rlimit['memory_size']),
                "limit_memory": str(rlimit['memory_size'])
            }
        )

        LOG.info(
            "Creating pod %s for image function:\n%s", pod_name, pod_body
        )

        try:
            self.v1.create_namespaced_pod(
                self.conf.kubernetes.namespace,
                body=yaml.safe_load(pod_body),
            )
        except Exception:
            LOG.exception("Failed to create pod.")
            raise exc.OrchestratorException('Execution preparation failed.')

    def _update_pod_label(self, pod, new_label):
        name = pod.metadata.name

        pod_labels = copy.deepcopy(pod.metadata.labels) or {}
        pod_labels.update(new_label)
        body = {
            'metadata': {
                'labels': pod_labels
            }
        }
        self.v1.patch_namespaced_pod(
            name, self.conf.kubernetes.namespace, body
        )

        LOG.debug('Labels updated for pod %s', name)

        return pod_labels

    def prepare_execution(self, function_id, version, rlimit=None, image=None,
                          identifier=None, labels=None, input=None):
        """Prepare service URL for function version.

        :param rlimit: optional argument passed to limit cpu/mem resources.

        For image function, create a single pod with rlimit and input, so the
        function will be executed in the resource limited pod.

        For normal function, choose a pod from the pool and expose a service,
        return the service URL.

        return a tuple includes pod name and the servise url.
        """
        pods = None

        labels = labels or {'function_id': function_id}

        if image:
            if not rlimit:
                LOG.critical('Param rlimit is required for image function.')
                raise exc.OrchestratorException(
                    'Execution preparation failed.'
                )

            self._create_pod(image, rlimit, identifier, labels, input)

            return identifier, None
        else:
            pods = self._choose_available_pods(labels, function_id=function_id,
                                               function_version=version)

        if not pods:
            LOG.critical('No worker available.')
            raise exc.OrchestratorException('Execution preparation failed.')

        try:
            pod_name, url = self._prepare_pod(
                pods[0], identifier, function_id, version, labels
            )
            return pod_name, url
        except Exception:
            LOG.exception('Pod preparation failed.')
            self.delete_function(function_id, version, labels)
            raise exc.OrchestratorException('Execution preparation failed.')

    def run_execution(self, execution_id, function_id, version, rlimit=None,
                      input=None, identifier=None, service_url=None,
                      entry='main.main', trust_id=None, timeout=None):
        """Run execution.

        Return a tuple including the result and the output.
        """
        if service_url:
            func_url = '%s/execute' % service_url
            data = utils.get_request_data(
                self.conf, function_id, version, execution_id, rlimit, input,
                entry, trust_id, self.qinling_endpoint, timeout
            )
            LOG.debug(
                'Invoke function %s(version %s), url: %s, data: %s',
                function_id, version, func_url, data
            )

            return utils.url_request(self.session, func_url, body=data)
        else:
            # Wait for image type function execution to be finished
            def _wait_complete():
                pod = self.v1.read_namespaced_pod(
                    identifier,
                    self.conf.kubernetes.namespace
                )
                status = pod.status.phase

                if status == 'Succeeded':
                    return pod

                raise exc.TimeoutException()

            duration = 0
            try:
                r = tenacity.Retrying(
                    wait=tenacity.wait_fixed(1),
                    stop=tenacity.stop_after_delay(timeout),
                    retry=tenacity.retry_if_exception_type(
                        exc.TimeoutException),
                    reraise=True
                )
                pod = r.call(_wait_complete)

                statuses = pod.status.container_statuses
                for s in statuses:
                    if hasattr(s.state, "terminated"):
                        end_time = s.state.terminated.finished_at
                        start_time = s.state.terminated.started_at
                        delta = end_time - start_time
                        duration = delta.seconds
                        break
            except exc.TimeoutException:
                LOG.exception(
                    "Timeout for function execution %s, pod %s",
                    execution_id, identifier
                )

                self.v1.delete_namespaced_pod(
                    identifier,
                    self.conf.kubernetes.namespace
                )
                LOG.debug('Pod %s deleted.', identifier)

                return False, {'output': 'Function execution timeout.',
                               'duration': timeout}
            except Exception:
                LOG.exception("Failed to wait for pod %s", identifier)
                return False, {'output': 'Function execution failed.',
                               'duration': duration}

            log = self.v1.read_namespaced_pod_log(
                identifier,
                self.conf.kubernetes.namespace,
            )

            return True, {'duration': duration, 'logs': log}

    def delete_function(self, function_id, version, labels=None):
        """Delete related resources for function.

        - Delete service
        - Delete pods
        """
        pre_label = {
            'function_id': function_id,
            'function_version': str(version)
        }
        labels = labels or pre_label
        selector = common.convert_dict_to_string(labels)

        ret = self.v1.list_namespaced_service(
            self.conf.kubernetes.namespace, label_selector=selector
        )
        names = [i.metadata.name for i in ret.items]
        for svc_name in names:
            self.v1.delete_namespaced_service(
                svc_name,
                self.conf.kubernetes.namespace
            )

        self.v1.delete_collection_namespaced_pod(
            self.conf.kubernetes.namespace,
            label_selector=selector
        )

    def scaleup_function(self, function_id, version, identifier=None, count=1):
        pod_names = []
        labels = {'runtime_id': identifier}
        pods = self._choose_available_pods(labels, count=count)

        if not pods:
            raise exc.OrchestratorException('Not enough workers available.')

        for pod in pods:
            pod_name, service_url = self._prepare_pod(
                pod, identifier, function_id, version, labels
            )
            pod_names.append(pod_name)

        LOG.info('Pods scaled up for function %s(version %s): %s', function_id,
                 version, pod_names)

        return pod_names, service_url

    def delete_worker(self, pod_name, **kwargs):
        self.v1.delete_namespaced_pod(
            pod_name,
            self.conf.kubernetes.namespace,
        )



# Copyright 2017 Catalyst IT Ltd
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from qinling_tempest_plugin.tests import base


class BasicOpsTest(base.BaseQinlingTest):
    name_prefix = 'BasicOpsTest'
    create_runtime = False

    @decorators.idempotent_id('205fd749-2468-4d9f-9c05-45558d6d8f9e')
    def test_basic_ops(self):
        """Basic qinling operations test case, including following steps:

        1. Admin user creates a runtime.
        2. Normal user creates function.
        3. Normal user creates execution(invoke function).
        4. Check result and execution log.
        """
        name = data_utils.rand_name('runtime', prefix=self.name_prefix)
        resp, body = self.admin_client.create_runtime(self.image, name)
        self.assertEqual(201, resp.status)
        self.assertEqual(name, body['name'])

        # Wait for runtime to be available
        self.runtime_id = body['id']
        self.wait_runtime_available(self.runtime_id)
        self.addCleanup(self.admin_client.delete_resource, 'runtimes',
                        self.runtime_id, ignore_notfound=True)

        # Create function
        function_id = self.create_function()

        # Invoke function
        resp, body = self.client.create_execution(
            function_id, input='{"name": "Qinling"}'
        )
        # self.assertEqual(201, resp.status)
        # self.assertEqual('success', body['status'])
        execution_id = body['id']
        self.addCleanup(self.client.delete_resource, 'executions',
                        execution_id, ignore_notfound=True)

        # Get execution log
        resp, body = self.client.get_execution_log(execution_id)

        self.assertEqual(200, resp.status)
        self.assertIn('Hello, Qinling', body)
